# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""API related to node creation in topology"""

import logging

from .address import Address
from .. import engine
from .id_generator import IdGen
from ..topology_map import TopologyMap

logger = logging.getLogger(__name__)


class Node:
    """
    Abstraction for a network namespace.

    Attributes
    ----------
    name: str
        User given name for the node.
    """

    def __init__(self, name):
        """
        Create a node with given `name`.

        An unique `id` is assigned to this node which is used by
        `engine` module to create the network namespace.
        This ensures that there is no naming conflict between any two
        nodes.

        Parameters
        ----------
        name: str
            The name of the node to be created
        """
        if name == '':
            raise ValueError('Node name can\'t be an empty string')

        self._name = name
        self._id = IdGen.get_id(name)
        self._interfaces = []

        engine.create_ns(self.id)
        TopologyMap.add_namespace(self.id, self.name)
        TopologyMap.add_host(self)

    def add_route(self, dest_addr, via_interface, next_hop_addr=''):
        """
        Add a route to the routing table of `Node`.

        Parameters
        ----------
        dest_addr: Address/str
            Destination ip address of node to route to. 'DEFAULT' is
            for all addresses
        via_interface: Interface
            `Interface` in `Node` to route via
        next_hop_addr: Address/str, optional
            IP address of next hop Node (or router), by default ''
        """
        if isinstance(dest_addr, str):
            dest_addr = Address(dest_addr)

        if next_hop_addr != '':
            if isinstance(next_hop_addr, str):
                next_hop_addr = Address(next_hop_addr)
        else:
            # Assuming veth pair
            next_hop_addr = via_interface.pair.address

        dest_addr_str = ''
        if dest_addr.is_subnet():
            dest_addr_str = dest_addr.get_addr()
        else:
            dest_addr_str = dest_addr.get_addr(with_subnet=False)

        engine.add_route(
            self.id, dest_addr_str, next_hop_addr.get_addr(with_subnet=False),
            via_interface.id)

    def _add_interface(self, interface):
        """
        Add `interface` to `Node`

        Parameters
        ----------
        interface: Interface
            `Interface` to be added to `Node`
        """
        self._interfaces.append(interface)
        interface.node = self
        engine.add_int_to_ns(self.id, interface.id)
        TopologyMap.add_interface(
            self.id, interface.id, interface.name)

    def configure_tcp_param(self, param, value):
        """
        Configure TCP parameters of `Node` available at
        /proc/sys/net/ipv4/tcp_*.

        Eg. 'window_scaling', 'wmem', 'ecn', etc.

        Parameters
        ----------
        param: str
            TCP parameter to be configured
        value: str
            New value of TCP parameter `param`

        Returns
        -------
        str
            If TCP Parameter `param` is valid, then new `value` is set
            for this `param`.
        """
        engine.configure_kernel_param(self.id, 'net.ipv4.tcp_', param, value)

    def configure_udp_param(self, param, value):
        """
        Configure UDP parameters of `Node` available at
        /proc/sys/net/ipv4/udp_*.

        Eg. 'early_demux', 'l3mdev_accept', 'rmem_min', 'wmem_min'

        Parameters
        ----------
        param: str
            TCP parameter to be configured
        value: str
            New value of TCP parameter `param`

        Returns
        -------
        str
            If TCP Parameter `param` is valid, then new `value` is set
            for this `param`.
        """
        engine.configure_kernel_param(self.id, 'net.ipv4.udp_', param, value)

    def read_tcp_param(self, param):
        """
        Read TCP parameters of `Node` available at
        `/proc/sys/net/ipv4/tcp_*`.

        Eg. 'window_scaling', 'wmem', 'ecn', etc.

        Parameters
        ----------
        param: str
            TCP parameter to be read

        Returns
        -------
        str
            If TCP Parameter `param` is valid, then corresponding value
            is returned.

        read tcp_parameters available at /proc/sys/net/ipv4/tcp_*
        Eg. window_scaling, wmem, ecn, etc.
        """
        return engine.read_kernel_param(self.id, 'net.ipv4.tcp_', param)

    def read_udp_param(self, param):
        """
        Read UDP parameters of `Node`available at
        `/proc/sys/net/ipv4/udp_*`.

        Eg. 'early_demux', 'l3mdev_accept', 'rmem_min', 'wmem_min'

        Parameters
        ----------
        param: str
            UDP parameter to be read

        Returns
        -------
        str
            If UDP Parameter `param` is valid, then corresponding value
            is returned.
        """
        return engine.read_kernel_param(self.id, 'net.ipv4.udp_', param)

    def ping(self, destination_address, verbose=True):
        """
        Ping from current `Node` to destination address
        if there is a route.

        Parameters
        ----------
        destination_address: Address/str
            IP address to ping to
        verbose: bool
            If `True`, print extensive ping success/failure details

        Returns
        -------
        bool
            `True` if `Node` can succesfully ping `destination_address`.
            Else `False`.
        """
        if isinstance(destination_address, str):
            destination_address = Address(destination_address)

        status = engine.ping(
            self.id, destination_address.get_addr(with_subnet=False))
        if verbose:
            if status:
                logger.debug('SUCCESS: ping from %s to %s', self.name,
                             destination_address.get_addr(with_subnet=False))
            else:
                logger.debug('FAILURE: ping from %s to %s', self.name,
                             destination_address.get_addr(with_subnet=False))
        return status

    def enable_ip_forwarding(self):
        """
        Enable IP forwarding in `Node`.

        After this method runs, the `Node` can be used as a router.
        """
        engine.en_ip_forwarding(self.id)
        TopologyMap.add_router(self)

    @property
    def id(self):
        """
        str: Value used by `engine` to create the emulated `Node` entity
        """
        return self._id

    @property
    def interfaces(self):
        """list(Interface): List of interfaces in this node"""
        return self._interfaces

    @property
    def name(self):
        """str: User given name for `Node`"""
        return self._name

    def __repr__(self):
        classname = self.__class__.__name__
        return f'{classname}({self.name!r})'



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""
Runs ss command and parses socket stats from
its output
"""

import os
import re
from ..results import SsResults
from .runnerbase import Runner
from ...topology_map import TopologyMap


class SsRunner(Runner):
    """
    Runs ss command and stores and parses the output

    Attributes
    ----------
    iterator : str
        absolute path of the ss iterator script
    param_list: list(str)
        list of parameters to be parsed
    ns_id : str
        network namespace to run ss from
    destination_ip : str
        ip address of the destination namespace
    start_time : num
        time at which ss is to be run
    run_time : num
        total time to run ss for
    """

    iterator = os.path.realpath(os.path.dirname(__file__)) + "/iterators/ss.sh"
    param_list = ['cwnd', 'rwnd', 'rtt', 'ssthresh',
                  'rto', 'delivery_rate', 'pacing_rate']

    def __init__(self, ns_id, destination_ip, start_time, run_time):
        """
        Constructor to initialize ss runner

        Parameters
        ----------
        ns_id : str
            network namespace to run ss from
        destination_ip : str
            ip address of the destination namespace
        start_time : num
            time at which ss is to be run
        run_time : num
            total time to run ss for
        """
        self.ns_id = ns_id
        self.destination_ip = destination_ip
        super().__init__(start_time, run_time)

    def run(self):
        """
        Runs the ss iterator
        """

        command = 'ip netns exec {ns_id} /bin/bash {iterator} {destination} {duration}\
            {filter}  {start_time}'.format(
                ns_id=self.ns_id, iterator=SsRunner.iterator, destination=self.destination_ip,
                duration=self.run_time, filter="\"dport != 12865 and sport != 12865\"",
                start_time=self.start_time)

        super().run(command)

    def print_error(self):
        """
        Method to print error from `self.err`
        """
        self.err.seek(0)  # rewind to start of file
        error = self.err.read().decode()
        ns_name = TopologyMap.get_namespace(self.ns_id)['name']
        self.logger.error('Collecting socket stats at %s. %s', ns_name, error)

    # pylint: disable=too-many-locals
    def parse(self):
        """
        parses the required data from `self.out`
        """
        stats_dict_list = {}

        self.out.seek(0)    # rewind to start of the temp file

        # See `iterators/ss.h` for output format
        raw_stats = self.out.read().decode().split("---")

        for raw_stat in raw_stats[:-1]:

            # Pattern to capture port numbers of flows to `destination ip`
            port_pattern = re.escape(self.destination_ip) + r':(?P<port>\d+)'
            port_list = [port.group('port')
                         for port in re.finditer(port_pattern, raw_stat)]
            timestamp_pattern = r'timestamp:(?P<timestamp>\d+\.\d+)'
            timestamp = re.search(
                timestamp_pattern, raw_stat).group("timestamp")

            for port in port_list:
                # If port encoutered first time
                if port not in stats_dict_list:
                    stats_dict_list[port] = [self.get_meta_item()]

                stats_dict_list[port].append({"timestamp": timestamp})

            for param in SsRunner.param_list:
                pattern = r'\s' + \
                    re.escape(param) + \
                    r'[\s:](?P<value>\w+\.?\w*(?:[\/\,]\w+\.?\w*)*)\s'
                # result list stores all the string that is matched by the `pattern`
                param_value_list = [value.group('value')
                                    for value in re.finditer(pattern, raw_stat)]
                param_value = ''
                for i in range(len(param_value_list)):
                    param_value = param_value_list[i].strip()
                    # remove the (rate) units at the end and convert
                    if param_value.endswith('bps'):
                        param_value = self.convert_to(param_value)
                    try:
                        # rtt has both avg and dev rtt separated by a /
                        if param == 'rtt':
                            avg_rtt = param_value.split('/')[0]
                            dev_rtt = param_value.split('/')[1]
                            stats_dict_list[port_list[i]][-1]['rtt'] = avg_rtt
                            stats_dict_list[port_list[i]
                                            ][-1]['dev_rtt'] = dev_rtt
                        else:
                            stats_dict_list[port_list[i]][-1][param] = param_value
                    except TypeError:
                        pass

        SsResults.add_result(
            self.ns_id, {self.destination_ip: stats_dict_list})
        self.clean_up()

    def clean_up(self):
        """
        Closes the temp files created
        """
        return super().clean_up()

    @staticmethod
    def convert_to(param_value, unit_out='Mbps'):
        """
        Converts parameter value to specified unit [For bit per second]
        """
        converter = {'bps': 1, 'Kbps': 1e3, 'Mbps': 1e6, 'Gbps': 1e9}

        # Extract value and unit
        unit_in = re.sub(r'[0-9]', '', param_value)
        extracted_param_value = float(re.sub(r'[A-Za-z]', '', param_value))

        converted_param_value = str(extracted_param_value * converter[unit_in]/converter[unit_out])
        return converted_param_value



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""
Base class for Quagga.
"""

from abc import ABC, abstractmethod
import io
import shutil
from nest.engine.quagga import chown_quagga


class QuaggaBase(ABC):
    """
    Abstract class for Quagga related processes.
    This class should be inherited for adding other quagga daemons.

    Use `add_to_config` to sequentially add daemon related commands
    for the config files.

    Call `create_config_command` to actually create the file on disk

    Finally call `run` to run the daemon. Uses the above created config file.

    Note:
    (ii): If you're using `RoutingHelper`, config files are created at the /tmp
          directory.
    (ii): Daemons may not work as expected if a config file is not created
    Attributes
    ----------
    conf: file-like
        In memory stream to temporarily hold configuration
    router_ns_id : str
        Router namespace id
    daemon : str
        quagga daemon to run(one of ['zebra', 'ospf'])
    conf_file : str
        config file path
    pid_file : str
        pid file path for the daemon process
    interfaces : List[Interface]
        interfaces present in the router
    """

    def __init__(self, router_ns_id, interfaces, daemon, conf_dir):
        self.conf = io.StringIO()
        self.router_ns_id = router_ns_id
        self.daemon = daemon
        self.conf_file = f'{conf_dir}/{self.router_ns_id}_{daemon}.conf'
        self.pid_file = f'{conf_dir}/{self.router_ns_id}_{daemon}.pid'
        self.interfaces = interfaces

    @abstractmethod
    def create_basic_config(self):
        """
        Created minimum configuration for `daemon`
        """

    @abstractmethod
    def run(self):
        """
        Run the `daemon` along with its config file
        """

    def add_to_config(self, command):
        """
        Add a line to `self.conf`

        Parmeters
        ---------
        command : str
            command to add to config file
        """
        self.conf.write(f'{command}\n')

    def create_config(self):
        """
        Creates config file on disk from `self.conf`
        """
        with open(self.conf_file, 'w') as conf:
            chown_quagga(self.conf_file)
            self.conf.seek(0)
            shutil.copyfileobj(self.conf, conf)



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Runs and rtt ouptut from ping command"""

import re
from time import sleep
from .runnerbase import Runner
from ..results import PingResults
from ...topology_map import TopologyMap


class PingRunner(Runner):
    """
    Runs ping command and parses statistics from it's output

    Attributes
    ----------
    ns_id : str
        network namespace to run netperf from
    destination_ip : str
        ip address of the destination namespace
    start_time : num
        time at which netperf is to run
    run_time : num
        total time to run netperf for
    """

    def __init__(self, ns_id, destination_ip, start_time, run_time):
        """
        Constructor to initialize ping runner

        Parameters
        ----------
        ns_id : str
            network namespace to run netperf from
        destination_ip : str
            ip address of the destination namespace
        start_time : num
            time at which netperf is to run
        run_time : num
            total time to run netperf for
        """
        self.ns_id = ns_id
        self.destination_ip = destination_ip
        super().__init__(start_time, run_time)

    # pylint: disable=arguments-differ
    def run(self):
        """
        Runs ping at t=`self.start_time`
        """
        if self.start_time > 0:
            sleep(self.start_time)
        command = f'ip netns exec {self.ns_id} ping {self.destination_ip} -w {self.run_time} -D \
            -i 0.2'
        super().run(command)

    def print_error(self):
        """
        Method to print error from `self.err`
        """
        self.err.seek(0)  # rewind to start of file
        error = self.err.read().decode()
        ns_name = TopologyMap.get_namespace(self.ns_id)['name']
        self.logger.error('Collecting latency at %s. %s', ns_name, error)

    def parse(self):
        """
        parses the rtt from `self.out`
        """
        self.out.seek(0)    # rewind to start of the temp file
        raw_stats = self.out.read().decode()

        pattern = r'\[(?P<timestamp>\d+\.\d+)\].*time=(?P<rtt>\d+(\.\d+)?)'
        timestamps_and_rtts = [(match.group('timestamp'), match.group(
            'rtt')) for match in re.finditer(pattern, raw_stats)]

        # List storing collected stats
        # First item as "meta" item with user given informataion
        stats_list = [self.get_meta_item()]

        for (timestamp, rtt) in timestamps_and_rtts:
            stats_list.append({
                'timestamp': timestamp,
                'rtt': rtt
            })

        stats_dict = {self.destination_ip: stats_list}

        PingResults.add_result(self.ns_id, stats_dict)

        self.clean_up()



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""quagga commands"""

from .exec import exec_subprocess


def run_zebra(ns_id, conf_file, pid_file):
    """
    Runs the zebra daemon

    Parameters
    ----------
    ns_id : str
        namespace of the router
    conf_file : str
        path to config file
    pid_file : str
        path to pid file
    """
    cmd = f'ip netns exec {ns_id} zebra --config_file {conf_file} \
            --pid_file {pid_file} --retain --daemon'
    exec_subprocess(cmd)


def run_ripd(ns_id, conf_file, pid_file):
    """
    Runs the zebra daemon

    Parameters
    ----------
    ns_id : str
        namespace of the router
    conf_file : str
        path to config file
    pid_file : str
        path to pid file
    """
    cmd = f'ip netns exec {ns_id} ripd --config_file {conf_file} \
            --pid_file {pid_file} --retain --daemon'
    exec_subprocess(cmd)


def run_ospfd(ns_id, conf_file, pid_file):
    """
    Runs the zebra daemon

    Parameters
    ----------
    ns_id : str
        namespace of the router
    conf_file : str
        path to config file
    pid_file : str
        path to pid file
    """
    cmd = f'ip netns exec {ns_id} ospfd --config_file {conf_file} --pid_file {pid_file} --daemon'
    exec_subprocess(cmd)


def chown_quagga(path):
    """
    Change ownership of quagga config directory and files to quagga

    Parameters
    ----------
    path : str
        path to file or directory
    """
    cmd = f'chown quagga {path}'
    exec_subprocess(cmd)



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Class to handles Ospf related functionalities"""

from nest.engine.quagga import run_ospfd
from nest.routing.quagga_base import QuaggaBase


class Ospf(QuaggaBase):
    """
    Handles Ospf related functionalities for quagga.
    """

    def __init__(self, router_ns_id, interfaces, conf_dir):
        super().__init__(router_ns_id, interfaces, 'ospfd', conf_dir)

    def create_basic_config(self):
        """
        Creates a file with basic configuration for ospf.
        Use base `add_to_config` directly for more complex configurations
        """
        for interface in self.interfaces:
            self.add_to_config(f'interface {interface.id}')

        self.add_to_config('router ospf')
        self.add_to_config(
            f'ospf router-id {self.interfaces[0].address.get_addr(with_subnet=False)}')
        for interface in self.interfaces:
            self.add_to_config(
                f' network {interface.address.get_subnet()} area 0.0.0.0')

        self.create_config()

    def run(self):
        """
        Runs the ospfd command
        """
        run_ospfd(self.router_ns_id, self.conf_file, self.pid_file)



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""iperf commands"""

from .exec import exec_subprocess


def run_iperf_server(ns_name):
    """
    Run Iperf Server on a namesapce

    Parameters
    ----------
    ns_name : str
        name of the server namespace
    """
    exec_subprocess(
        f'ip netns exec {ns_name}  iperf3 -s -D')   # runs server as a daemon


def run_iperf_client(ns_name, server_ip, run_time, flows, target_bw):
    """
    Run Iperf Client

    Parameters
    ----------
    ns_name : str
        name of the client namespace
    server_ip : str
        the ip of server to which it has to connect
    run_time : num
        test duration
    flows : int
        number of parallel flows
    target_bw : int
        target bandwidth of the udp flow in mbits
    """

    exec_subprocess(
        f'ip netns exec {ns_name} iperf3 -u -c {server_ip} -P {flows} -t {run_time} \
            -b {target_bw}m -i 0')



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Package all results into a folder"""

import os
import time

from nest.user import User


class Pack():
    """Handles packaging results"""

    FOLDER = ''

    @staticmethod
    def init(exp_name):
        """
        Create a folder with format 'testname(time)_dump'

        Parameters
        ----------
        exp_name : str
            Name of experiment
        """
        timestamp = time.strftime('%d-%m-%Y-%H:%M:%S')
        Pack.FOLDER = '{exp_name}({timestamp})_dump'.format(
            exp_name=exp_name, timestamp=timestamp)
        os.mkdir(Pack.FOLDER)
        Pack.set_owner(Pack.FOLDER)

    @staticmethod
    def dump_file(filename, content):
        """
        Dump a file into Pack.FOLDER

        Parameters
        ----------
        filename : str
            Name of file
        content : str
            Content to be stored in file
        """
        path = os.path.join(Pack.FOLDER, filename)
        with open(path, 'w') as file:
            file.write(content)
        Pack.set_owner(path)

    @staticmethod
    def dump_plot(subfolder, filename, fig):
        """
        Dump a plot into Pack.FOLDER

        Parameters
        ----------
        subfolder : str
            Subfolder to which plot belongs to
        filename : str
            Name of plot
        fig : matplotlib.pyplot.fig
            Plot figure to be stored in file
        """
        Pack.create_subfolder(subfolder)
        path = os.path.join(Pack.FOLDER, subfolder, filename)
        fig.savefig(path)
        Pack.set_owner(path)

    @staticmethod
    def create_subfolder(subfolder):
        """
        Create subfolder if it already doesn't exist

        Parameters
        ----------
        subfolder : str
            Subfolder name
        """
        path = os.path.join(Pack.FOLDER, subfolder)
        if not os.path.isdir(path):
            os.mkdir(path)
            Pack.set_owner(path)

    @staticmethod
    def set_owner(path):
        """
        Change file permission to that of the
        actual user running (For eg. user running
        sudo)

        Parameters
        ----------
        path : str
            Path of file
        """

        user_id = User.user_id
        group_id = User.group_id
        os.chown(path, user_id, group_id)

    @staticmethod
    def compress():
        """Compress Pack.FOLDER into a tar archive"""
        # TODO



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

""" Runs iperf command to setup udp flows """

from time import sleep
from .runnerbase import Runner
from ...engine.iperf import run_iperf_server, run_iperf_client

# NOTE: Uses iperf3 command
class IperfRunner(Runner):
    """
    Runs iperf client and server. Currently being used
    for udp flows only

    Attributes
    ----------
    ns_id : str
        network namespace to run iperf from
    """

    # pylint: disable=too-many-arguments
    def __init__(self, ns_id, destination_ip, bandwidth, n_flows, start_time, run_time):
        """
        Constructor to initalize the runner

        Parameters
        ----------
        ns_id : str
            network namespace to run iperf from
        destination : str
            the ip of server to which it has to connect
        bandwidth : int
            target bandwidth of the udp flow in mbits
        n_flows : int
            number of parallel flows
        start_time : num
            start time of the flow
        run_time : num
            test duration
        """
        self.ns_id = ns_id
        self.destination_ip = destination_ip
        self.bandwidth = bandwidth
        self.n_flows = n_flows
        super().__init__(start_time, run_time)

    # Should this be placed somewhere else?
    @staticmethod
    def run_server(ns_id):
        """
        Run iperf server in `ns_id`

        Parameters
        ----------
        ns_id : str
            namespace to run netserver on
        """
        run_iperf_server(ns_id)

    # This method should call super().run
    def run(self):
        """
        calls engine method to run iperf client
        """
        if self.start_time != 0:
            sleep(self.start_time)
        run_iperf_client(self.ns_id, self.destination_ip,
                         self.run_time, self.n_flows, self.bandwidth)



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""
This module contains methods to kill any running processes in namespaces
and delete all namespaces after the experiment is complete.
"""

import atexit
import logging
from . import config

from . import engine
from .topology_map import TopologyMap

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def kill_processes():
    """
    Kill any running processes in namespaces
    """
    for namespace in TopologyMap.get_namespaces():
        engine.kill_all_processes(namespace['id'])


@atexit.register
def delete_namespaces():
    """
    Delete all the newly generated namespaces
    """
    namespaces = TopologyMap.get_namespaces()

    if config.get_value('delete_namespaces_on_termination'):
        for namepspace in namespaces:
            engine.delete_ns(namepspace['id'])
        logger.info('Cleaned up environment!')
    else:
        logger.info('Namespaces not deleted')



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Test APIs from topology sub-package"""

import unittest
from nest.topology import Node, connect
from nest.clean_up import delete_namespaces

#pylint: disable=missing-docstring
#pylint: disable=invalid-name
class TestTopology(unittest.TestCase):

    def setUp(self):
        self.n0 = Node('n0')
        self.n1 = Node('n1')

    def tearDown(self):
        delete_namespaces()


    def test_p2p(self):
        (n0_n1, n1_n0) = connect(self.n0, self.n1)

        n0_n1.set_address('10.0.0.1/24')
        n1_n0.set_address('10.0.0.2/24')

        status = self.n0.ping('10.0.0.2', verbose=False)

        self.assertTrue(status)

    def test_prp(self):
        # pylint: disable=invalid-name
        r = Node('r')
        r.enable_ip_forwarding()

        (n0_r, r_n0) = connect(self.n0, r)
        (r_n1, n1_r) = connect(r, self.n1)

        n0_r.set_address('10.1.1.1/24')
        r_n0.set_address('10.1.1.2/24')
        r_n1.set_address('10.1.2.2/24')
        n1_r.set_address('10.1.2.1/24')

        self.n0.add_route('DEFAULT', n0_r)
        self.n1.add_route('DEFAULT', n1_r)

        status = self.n0.ping('10.1.2.1', verbose=False)

        self.assertTrue(status)

    def test_tcp_param(self):
        self.n0.configure_tcp_param('ecn', '1')
        ecn = self.n0.read_tcp_param('ecn')
        self.assertEqual(ecn, '1')

if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Sysctl commands"""

from .exec import exec_subprocess

def en_ip_forwarding(ns_name):
    """
    Enables ip forwarding in a namespace. Used for routers

    Parameters
    ----------
    ns_name : str
        namespace name
    """
    exec_subprocess(f'ip netns exec {ns_name} sysctl -w net.ipv4.ip_forward=1')

def configure_kernel_param(ns_name, prefix, param, value):
    """
    Configure kernel parameters using sysctl

    Parameters
    ----------
    ns_name : str
        name of the namespace
    prefix : str
        path for the sysctl command
    param : str
        kernel parameter to be configured
    value : str
        value of the parameter
    """
    exec_subprocess(f'ip netns exec {ns_name} sysctl -q -w {prefix}{param}={value}')


def read_kernel_param(ns_name, prefix, param):
    """
    Read kernel parameters using sysctl

    Parameters
    ----------
    ns_name : str
        name of the namespace
    prefix : str
        path for the sysctl command
    param : str
        kernel parameter to be read

    Returns
    -------
    str
        value of the `param`
    """
    value = exec_subprocess(f'ip netns exec {ns_name} sysctl -n {prefix}{param}', output=True)
    return value.rstrip('\n')



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Test APIs from routing sub-package"""

import unittest
from nest.topology import Node, connect
from nest.routing.routing_helper import RoutingHelper
from nest.clean_up import delete_namespaces

#pylint: disable=missing-docstring

class TestQuagga(unittest.TestCase):

    # pylint: disable=invalid-name
    @classmethod
    def setUpClass(cls):
        cls.n0 = Node('n0')
        cls.n1 = Node('n1')
        cls.r0 = Node('r0')
        cls.r1 = Node('r1')
        cls.r0.enable_ip_forwarding()
        cls.r1.enable_ip_forwarding()

        ### Create interfaces and connect nodes and routers ###

        (eth_p1r1, eth_r1p1) = connect(cls.n0, cls.r0, 'eth-n1r1-0', 'eth-r1n1-0')
        (eth_r1r2, eth_r2r1) = connect(cls.r0, cls.r1, 'eth-r1r2-0', 'eth-r2r1-0')
        (eth_r2p2, eth_p2r2) = connect(cls.r1, cls.n1, 'eth-r2n2-0', 'eth-n2r2-0')

        ### Assign addresses to interfaces ###

        eth_p1r1.set_address('10.0.1.1/24')
        eth_r1p1.set_address('10.0.1.2/24')

        eth_r1r2.set_address('10.0.2.2/24')
        eth_r2r1.set_address('10.0.2.3/24')

        eth_r2p2.set_address('10.0.3.3/24')
        eth_p2r2.set_address('10.0.3.4/24')

    @classmethod
    def tearDownClass(cls):
        delete_namespaces()


    def test_routing_helper(self):

        RoutingHelper('rip').populate_routing_tables()

        status = self.n0.ping('10.0.3.4', verbose=False)
        self.assertTrue(status)

        status = self.n1.ping('10.0.1.1', verbose=False)
        self.assertTrue(status)


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""
NeST: Network Stack Tester
==========================

This is the entry point of NeST package.

Following actions are performed as part of setup, when
nest is imported:
1. Check if nest is run with root privileges
2. Store SUDO user and group id information
"""

import logging
import os
import sys

from .logging_helper import add_logging_level
from .user import User
from . import config

if os.geteuid() != 0:
    print('nest: python package requires root access', file=sys.stderr)
    sys.exit(1)

# Load default config values
config.default_value = config.import_default_config()

# Store user information for later use
if all(key in os.environ for key in ('SUDO_UID', 'SUDO_GID')):
    user_id = int(os.environ['SUDO_UID'])
    group_id = int(os.environ['SUDO_GID'])
    User(user_id, group_id)

# Set up logging
# Added to avoid exec commands being printed everytime
add_logging_level('TRACE', logging.DEBUG - 1, 'trace')
logger = logging.getLogger(__name__)
ch = logging.StreamHandler()    # Logger output will be output to stderr
ch.setLevel(config.get_value('log_level'))
logger.addHandler(ch)
formatter = logging.Formatter('[%(levelname)s] : %(message)s')
ch.setFormatter(formatter)



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""
Helper class for routing
"""

import time
import logging
import importlib
from os import mkdir, kill, path
import atexit
from shutil import rmtree
from signal import SIGTERM
from nest.topology.id_generator import IdGen
from nest.routing.zebra import Zebra
from nest.topology_map import TopologyMap
from nest.engine.quagga import chown_quagga

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# pylint:disable=too-few-public-methods


class RoutingHelper:
    """
    Handles basic routing requirements for the topology.
    Either inherit this class or use `Zebra` and other protocols
    for better customizations

    Attributes
    ----------
    _module_map : dict
        map between protocol string and its module and class
    protocol : str
        routing protocol(one of['ospf'])
    routers : List[Node]
        routers in the topology
    hosts : List[Node]
        hosts in the topology
    conf_dir : str
        path for quagga config directory
    protocol_class : ABCMeta
        Protocol class which will later be instantiated
    """

    _module_map = {
        'ospf': ['nest.routing.ospf', 'Ospf'],
        'rip': ['nest.routing.rip', 'Rip']
    }

    def __init__(self, protocol='static', hosts=None, routers=None):
        """
        Constructor for RoutingHelper

        Parameters
        ----------
        protocol: str
            routing protocol to be run. one of [ospf, rip]
        hosts: List[Node]
            List of hosts in the network. If `None`, considers the entire topology.
            Use this if your topology has disjoint networks
        routers: List[Node]
            List of routers in the network. If `None`, considers the entire topology.
            Use this if your topology has disjoint networks
        """
        if protocol == 'static':
            raise NotImplementedError(
                'Static routing is yet to be implemented. Use rip or ospf')
        self.protocol = protocol
        self.routers = TopologyMap.get_routers() if routers is None else routers
        self.hosts = TopologyMap.get_hosts() if hosts is None else hosts
        self.conf_dir = None
        module_str, class_str = RoutingHelper._module_map[self.protocol]
        module = importlib.import_module(module_str)
        self.protocol_class = getattr(module, class_str)
        self.zebra_list = []
        self.protocol_list = []

        atexit.register(self._clean_up)

    def populate_routing_tables(self):
        """
        Populate routing table using `self.protocol`
        """
        self._setup_default_routes()

        if self.protocol == 'static':
            pass  # TODO: add static routing
        else:
            self._run_quagga()

    def _create_quagga_directory(self):
        """
        Creates a directly for holding quagga related config
        and pid files.
        Override this to create directory at a location other than /tmp

        Returns
        -------
        str:
            path of the created directory
        """
        dir_path = f'/tmp/quagga-configs_{IdGen.topology_id}'
        mkdir(dir_path)
        chown_quagga(dir_path)
        return dir_path

    def _setup_default_routes(self):
        """
        Setup default routes in hosts
        """
        for host in self.hosts:
            host.add_route('DEFAULT', host.interfaces[0])

    def _run_quagga(self):
        """
        Run zebra and `self.protocol`
        """
        logger.info('Running zebra and %s on routers', self.protocol)
        self.conf_dir = self._create_quagga_directory()
        for router in self.routers:
            self._run_zebra(router)
            self._run_routing_protocol(router)
        self._check_for_convergence()

    def _run_zebra(self, router):
        """
        Create requird config file and run zebra
        """
        zebra = Zebra(router.id, router.interfaces, self.conf_dir)
        zebra.create_basic_config()
        zebra.run()
        self.zebra_list.append(zebra)

    def _run_routing_protocol(self, router):
        """
        Create requird config file and run `self.protocol`
        """
        protocol = self.protocol_class(
            router.id, router.interfaces, self.conf_dir)
        protocol.create_basic_config()
        protocol.run()
        self.protocol_list.append(protocol)

    def _check_for_convergence(self):
        """
        Wait for the routing protocol to converge.
        Override this for custom convergence check
        """
        logger.info('Waiting for %s to converge', self.protocol)
        interval = 2
        converged = False
        # Ping between hosts until convergence
        while not converged:
            time.sleep(interval)
            converged = True
            for i in range(len(self.hosts)):
                for j in range(i+1, len(self.hosts)):
                    if not self.hosts[i].ping(self.hosts[j].interfaces[0].address.get_addr(),
                                              verbose=False):
                        converged = False
                        break
                if not converged:
                    break

        logger.info('Routing compeleted')
        self._clean_up()

    def _clean_up(self):
        """
        Terminates quagga daemons and deletes config files
        """

        # Stop zebra processes
        for zebra in self.zebra_list:
            if path.isfile(zebra.pid_file):
                with open(zebra.pid_file, 'r') as pid_file:
                    pid = int(pid_file.read())
                kill(pid, SIGTERM)

        # Stop protocol processes
        for protocol in self.protocol_list:
            if path.isfile(protocol.pid_file):
                with open(protocol.pid_file, 'r') as pid_file:
                    pid = int(pid_file.read())
                kill(pid, SIGTERM)

        # Delete config directory
        if path.isdir(self.conf_dir):
            rmtree(self.conf_dir)



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Ping command"""

from .exec import exec_subprocess

def ping(ns_name, dest_addr):
    """
    Send a ping packet from ns_name to dest_addr
    if possible

    Parameters
    ----------
    ns_name : str
        namespace name
    dest_addr : str
        address to ping to

    Returns
    -------
    bool
        success of ping
    """
    status = exec_subprocess(f'ip netns exec {ns_name} ping -c1 -q {dest_addr}')
    return status == 0



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

""" API's to process config files """

import os
import json
import logging

logger = logging.getLogger(__name__)

default_value = {'topology': {}, 'experiment': {}, 'engine': {}}

def import_default_config():
    """
    Reads `config.json` file from necessary specified file locations and converts it into a string

    Returns
    -------
    dict
        The information in the default config file as a python dictionary
    """
    with open(os.path.realpath(os.path.dirname(__file__)) + '/config.json', 'r') as json_file:
        data = json.load(json_file)
        return data

def set_value(parameter, value):
    """
    Changes the default values only for that program

    Attributes
    ----------
    parameter: str
        The parameter's value which has to be changed
    value: str
        The value to which the parameter has to be changed to
    """
    if parameter in default_value:
        default_value[parameter] = value
    elif parameter in default_value['topology']:
        default_value['topology'][parameter] = value
    elif parameter in default_value['experiment']:
        default_value['experiment'][parameter] = value
    elif parameter in default_value['engine']:
        default_value['engine'][parameter] = value
    else:
        logging.error('The given parameter %s does not exist', parameter)

def get_value(parameter):
    """
    Returns the default value of the parameter

    Attributes
    ----------
    parameter: str
        The parameter whose value is to be returned

    Returns
    -------
    str
        The value of the parameter
    """
    if parameter in default_value:
        return default_value[parameter]
    if parameter in default_value['topology']:
        return default_value['topology'][parameter]
    if parameter in default_value['experiment']:
        return default_value['experiment'][parameter]
    if parameter in default_value['engine']:
        return default_value['engine'][parameter]
    # If it belonged to none of them
    logging.error('The given parameter %s does not exist', parameter)
    return 'The given parameter %s does not exist'

def import_custom_config(path):
    """
    Overwrites the current value of config parameters with the values
    from the JSON in the given path

    Attributes
    ----------
    path: str
        The path in which the json file
    """
    with open(path, 'r') as json_file:
        data = json.load(json_file)
    if 'topology' in data:
        for parameter in data['topology']:
            if parameter in default_value['topology']:
                default_value['topology'][parameter] = data['topology'][parameter]
    elif 'experiment' in data:
        for parameter in data['experiment']:
            if parameter in default_value['experiment']:
                default_value['experiment'][parameter] = data['experiment'][parameter]
    elif 'engine' in data:
        for parameter in data['engine']:
            if parameter in default_value['engine']:
                default_value['engine'][parameter] = data['engine'][parameter]
    for parameter in data:
        if not isinstance(default_value[parameter], dict):
            default_value[parameter] = data[parameter]



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""
Runs netperf commands to setup tcp flow
and collect throughput data
"""

import re
import copy
from ..results import NetperfResults
from ...engine import exec_exp_commands
from .runnerbase import Runner
from ...topology_map import TopologyMap


class NetperfRunner(Runner):
    """
    Runs netperf command and parses statistics from it's output

    Attributes
    ----------
    default_netperf_options : dict
        default options to run netperf command with
    netperf_tcp_options : dict
        tcp related netperf options
    netperf_udp_options : dict
        udp related netperf options
    ns_id : str
        network namespace to run netperf from
    destination_ip : str
        ip address of the destination namespace
    start_time : num
        time at which netperf is to run
    run_time : num
        total time to run netperf for
    """

    tcp_output_options = [
        'THROUGHPUT', 'LOCAL_CONG_CONTROL', 'REMOTE_CONG_CONTROL', 'TRANSPORT_MSS',
        'LOCAL_SOCKET_TOS', 'REMOTE_SOCKET_TOS'
    ]

    default_netperf_options = {
        'banner': '-P 0',                   # Disable test banner
        'ipv4': '-4',                       # IPv4 Addresses
        'testname': '-t TCP_STREAM',        # Test type (NOTE: TCP_STREAM only for now)
        'fill_file': '-F /dev/urandom',     # File to transmit (NOTE: Inspired from flent)
        'testlen': '-l 10',                 # Length of test (NOTE: Default 10s)
        'intervel': '-D -0.2',              # Generated interim results every INTERVAL secs
        'debug': '-d',                      # Enable debug mode
    }

    netperf_tcp_options = {
        'cong_algo': '-K cubic',             # Congestion algorithm
        'stats': '-k THROUGHPUT'             # Stats required
    }

    netperf_udp_options = {
        'routing': '-R 1',                    # Enable routing
        'stats': '-k THROUGHPUT'              # Stats required
    }

    def __init__(self, ns_id, destination_ip, start_time, run_time, **kwargs):
        """
        Constructor to initialize netperf runner

        Parameters
        ----------
        ns_id : str
            network namespace to run netperf from
        destination_ip : str
            ip address of the destination namespace
        start_time : num
            time at which netperf is to run
        run_time : num
            total time to run netperf for
        **kwargs
            netperf options to override
        """
        self.ns_id = ns_id
        self.destination_ip = destination_ip
        self.options = copy.deepcopy(kwargs)
        super().__init__(start_time, run_time)

    # Should this be placed somewhere else?
    @staticmethod
    def run_netserver(ns_id):
        """
        Run netserver in `ns_id`

        Parameters
        ----------
        ns_id : str
            namespace to run netserver on
        """
        command = f'ip netns exec {ns_id} netserver'
        return_code = exec_exp_commands(command)
        if return_code != 0:
            ns_name = TopologyMap.get_namespace(ns_id)['name']
            print("Error running netserver at {}.".format(
                ns_name))

    def run(self):
        """
        Runs netperf at t=`self.start_time`
        """
        netperf_options = copy.copy(NetperfRunner.default_netperf_options)
        test_options = None

        # Change the default runtime
        netperf_options['testlen'] = '-l {}'.format(self.run_time)
        # Set test
        netperf_options['testname'] = '-t {}'.format(self.options['testname'])

        if netperf_options['testname'] == '-t TCP_STREAM':
            test_options = copy.copy(NetperfRunner.netperf_tcp_options)
            test_options['cong_alg'] = '-K {}'.format(
                self.options['cong_algo'])
        elif netperf_options['testname'] == '-t UDP_STREAM':
            test_options = copy.copy(NetperfRunner.netperf_udp_options)

        netperf_options_list = list(netperf_options.values())
        netperf_options_string = ' '.join(netperf_options_list)
        test_options_list = list(test_options.values())
        test_options_string = ' '.join(test_options_list)

        command = f'ip netns exec {self.ns_id} netperf {netperf_options_string} -H \
            {self.destination_ip} -s {self.start_time} -- {test_options_string}'

        super().run(command)

    def print_error(self):
        """
        Method to print error from `self.err`
        """
        self.err.seek(0)  # rewind to start of file
        error = self.err.read().decode()
        ns_name = TopologyMap.get_namespace(self.ns_id)['name']
        self.logger.error('Running netperf at %s. %s', ns_name, error)

    def parse(self):
        """
        Parse netperf output from `self.out`
        """
        self.out.seek(0)    # rewind to start of the temp file
        raw_stats = self.out.read().decode()

        # pattern that matches the netperf output corresponding to throughput
        throughput_pattern = r'NETPERF_INTERIM_RESULT\[\d+]=(?P<throughput>\d+\.\d+)'
        throughputs = [throughput.group('throughput') for throughput in re.finditer(
            throughput_pattern, raw_stats)]

        # pattern that matches the netperf output corresponding to interval
        timestamp_pattern = r'NETPERF_ENDING\[\d+]=(?P<timestamp>\d+\.\d+)'
        timestamps = [timestamp.group('timestamp') for timestamp in re.finditer(
            timestamp_pattern, raw_stats)]

        # pattern that gives the remote port
        remote_port_pattern = r'remote port is (?P<remote>\d+)'
        remote_port = re.search(remote_port_pattern, raw_stats).group('remote')

        # List storing collected stats
        # First item as "meta" item with user given informataion
        stats_list = [self.get_meta_item()]

        for i in range(len(throughputs)):
            stats_list.append({
                'timestamp': timestamps[i],
                # Netperf provides throughput as sending rate from sender's side
                'sending_rate': throughputs[i]
            })

        stats_dict = {f'{self.destination_ip}:{remote_port}': stats_list}

        NetperfResults.add_result(self.ns_id, stats_dict)
        self.clean_up()

    def clean_up(self):
        """
        Closes the temp files created
        """
        return super().clean_up()



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Test Address sub-module in topology"""

import unittest
from nest.topology.address import Address, Subnet

#pylint: disable=missing-docstring

class TestTopologyAddress(unittest.TestCase):

    def setUp(self):
        self.addr = '10.0.0.1/24'
        self.addr_without_subnet = '10.0.0.1'
        self.default_addr = 'DEFAULT'
        self.subnet_addr = '10.0.0.0/24'

    def test_get_addr(self):
        address = Address(self.addr)
        self.assertEqual(address.get_addr(), self.addr)
        self.assertEqual(address.get_addr(with_subnet=False), self.addr_without_subnet)

        with self.assertRaises(ValueError):
            Address('10.0.0.0a/24') #invalid character 'a'

        with self.assertRaises(ValueError):
            Address('10.0asdfasf') #invalid string

        with self.assertRaises(ValueError):
            Address('10.0.0./24') #invalid ./

        with self.assertRaises(ValueError):
            Address('10.0.0.2000/24') #invalid mask 2000

        with self.assertRaises(ValueError):
            Address('256.0.0.0/24') #invalid mask 256

    def test_address_subnet(self):
        subnet_address = Address(self.subnet_addr)
        self.assertTrue(subnet_address.is_subnet())

        address = Address(self.addr)
        self.assertEqual(address.get_subnet(), self.subnet_addr)

        with self.assertRaises(ValueError):
            Subnet('10.0.0.1/24') #10.0.0.1/24 is not a subnet


    def test_subnet(self):
        subnet = Subnet(self.subnet_addr)
        self.assertEqual(subnet.get_next_addr().get_addr(), '10.0.0.1/24')
        self.assertEqual(subnet.get_next_addr().get_addr(), '10.0.0.2/24')
        self.assertEqual(subnet.get_next_addr().get_addr(), '10.0.0.3/24')

        with self.assertRaises(ValueError):
            Subnet('10.0.0.1/24').get_next_addr()

        with self.assertRaises(ValueError):
            Subnet('10.0.0.0/32').get_next_addr() #no next address possible as the mask is 32

    def test_default(self):
        def_addr = Address('DEFAULT')
        self.assertEqual('default', def_addr.get_addr())
        self.assertEqual(False, def_addr.is_subnet())

        with self.assertRaises(Exception):
            def_addr.get_subnet() #no subnet for default

        with self.assertRaises(Exception):
            Subnet('DEFAULT') #DEFAULT cannot be a subnet

if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Traffic control management"""

from .exec import exec_subprocess

# Only bandwith and latency is considered
# Assuming tc on egress
# Using Netem

def add_traffic_control(host_name, dev_name, rate, latency):
    """
    Add traffic control to host

    Parameters
    ----------
    host_name : str
        name of the host namespace
    rate : str
        rate of the bandwidth
    latency : str
        latency of the link
    dev_name : str
    """
    exec_subprocess(f'tc -n {host_name} qdisc add dev {dev_name} root'
                    f' netem rate {rate} latency {latency}')


def add_qdisc(ns_name, dev_name, qdisc, parent='', handle='', **kwargs):
    """
    Add a qdisc on an interface

    Parameters
    ----------
    ns_name : str
        name of the namespace
    qdisc : str
        qdisc used on the interface
    parent : str
        id of the parent class in major:minor form(optional) (Default value = '')
    handle : str
        id of the qdisc in major:0 form (Default value = '')
    dev_name : str
        name of the interface
    """
    if parent and parent != 'root':
        parent = 'parent ' + parent

    if handle:
        handle = 'handle ' + handle

    qdisc_params = ''
    for param, value in kwargs.items():
        qdisc_params += param + ' ' + value + ' '

    exec_subprocess(f'tc -n {ns_name} qdisc add dev {dev_name}'
                    f' {parent} {handle} {qdisc} {qdisc_params}')


def change_qdisc(ns_name, dev_name, qdisc, parent='', handle='', **kwargs):
    """
    Change a qdisc that is already present on an interface

    Parameters
    ----------
    ns_name : str
        name of the namespace
    qdisc : str
        qdisc used on the interface
    parent : str
        id of the parent class in major:minor form(optional) (Default value = '')
    handle : str
        id of the qdisc in major:0 form (Default value = '')
    dev_name : str
        name of the interface
    """
    if parent and parent != 'root':
        parent = 'parent ' + parent

    if handle:
        handle = 'handle ' + handle

    qdisc_params = ''
    for param, value in kwargs.items():
        qdisc_params += param + ' ' + value + ' '

    exec_subprocess(f'tc -n {ns_name} qdisc change dev {dev_name}'
                    f' {parent} {handle} {qdisc} {qdisc_params}')


def replace_qdisc(ns_name, dev_name, qdisc, parent='', handle='', **kwargs):
    """
    Replace a qdisc that is already present on an interface

    Parameters
    ----------
    ns_name : str
        name of the namespace
    qdisc : str
        qdisc used on the interface
    parent : str
        id of the parent class in major:minor form(optional) (Default value = '')
    handle : str
        id of the qdisc in major:0 form (Default value = '')
    dev_name : str
        name of the interface
    """
    if parent and parent != 'root':
        parent = 'parent ' + parent

    if handle:
        handle = 'handle ' + handle

    qdisc_params = ''
    for param, value in kwargs.items():
        qdisc_params += param + ' ' + value + ' '

    exec_subprocess(f'tc -n {ns_name} qdisc replace dev {dev_name}'
                    f' {parent} {handle} {qdisc} {qdisc_params}')


def delete_qdisc(ns_name, dev_name, parent='', handle=''):
    """
    Add a qdisc on an interface

    Parameters
    ----------
    ns_name : str
        name of the namespace
    qdisc : str
        qdisc used on the interface
    parent : str
        id of the parent class in major:minor form(optional) (Default value = '')
    handle : str
        id of the qdisc in major:0 form (Default value = '')
    dev_name : str
        name of the interface
    """
    if parent and parent != 'root':
        parent = 'parent ' + parent

    if handle:
        handle = 'handle ' + handle

    exec_subprocess(f'tc -n {ns_name} qdisc del dev {dev_name} {parent} {handle}')


def add_class(ns_name, dev_name, parent, qdisc, classid='', **kwargs):
    """
    Add a class to a qdisc

    Parameters
    ----------
    ns_name : str
        name of the namespace
    parent : str
        id of the parent class in major:minor form(optional)
    qdisc : str
        qdisc used on the interface
    classid : str
        id of the class in major:minor form (Default value = '')
    dev_name : str
        name of the interface
    """
    if classid:
        classid = 'classid ' + classid

    qdisc_params = ''
    for param, value in kwargs.items():
        qdisc_params += param + ' ' + value + ' '

    exec_subprocess(f'tc -n {ns_name} class add dev {dev_name} parent {parent}'
                    f' {classid} {qdisc} {qdisc_params}')


def change_class(ns_name, dev_name, parent, qdisc, classid='', **kwargs):
    """
    Change a class that is already present on an interface

    Parameters
    ----------
    ns_name : str
        name of the namespace
    parent : str
        id of the parent class in major:minor form(optional)
    qdisc : str
        qdisc used on the interface
    classid : str
        id of the class in major:minor form (Default value = '')
    dev_name : str
        name of the interface
    """

    if classid:
        classid = 'classid ' + classid

    qdisc_params = ''
    for param, value in kwargs.items():
        qdisc_params += param + ' ' + value + ' '

    exec_subprocess(f'tc -n {ns_name} class change dev {dev_name} parent {parent}'
                    f' {classid} {qdisc} {qdisc_params}')


# pylint: disable=too-many-arguments
def add_filter(ns_name, dev_name, protocol, priority, filtertype, parent='', handle='', **kwargs):
    """
    Add a filter to a class

    Parameters
    ----------
    ns_name : str
        name of the namespace
    protocol : str
        protocol used
    priority : str
        priority
    filtertype : str
        one of the available filters
    parent : str
        id of the parent class in major:minor form(optional) (Default value = '')
    handle : str
        id of the filter (Default value = '')
    qdisc : str
        qdisc used on the interface
    dev_name : str
        name of the interface
    """

    # TODO: Check if protocol can be removed from the arguments since it's always ip

    if parent and parent != 'root':
        parent = 'parent ' + parent

    if handle:
        handle = 'handle ' + handle

    filter_params = ''

    for param, value in kwargs.items():
        filter_params += param + ' ' + value + ' '

    exec_subprocess(f'tc -n {ns_name} filter add dev {dev_name} {parent} {handle}'
                    f' protocol {protocol} prio {priority} {filtertype} {filter_params}')

def get_tc_version():
    """
    check for current tc version

    Returns
    -------
    str
        string containing current tc version
    """
    return exec_subprocess('tc -V', output=True)



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Class to handles RIP related functionalities"""

from nest.engine.quagga import run_ripd
from nest.routing.quagga_base import QuaggaBase


class Rip(QuaggaBase):
    """
    Handles RIP related functionalities for quagga.
    Refer to `QuaggaBase` for usage
    """

    def __init__(self, router_ns_id, interfaces, conf_dir):
        super().__init__(router_ns_id, interfaces, 'ripd', conf_dir)

    def add_rip(self):
        """
        Add command to enable rip on router to config file
        """
        self.add_to_config('router rip')

    def add_version(self, version):
        """
        Add version command to config file

        Parameters
        ----------
        version : str
            rip version. Recommended is v2
        """
        self.add_to_config(f' version {version}')

    def add_network(self, network):
        """
        Add command for subnet or interface to run rip on to config file

        Parameters
        ----------
        network : str
            subnet or interface to run rip on
        """
        self.add_to_config(f' network {network}')

    def create_basic_config(self):
        """
        Creates a file with basic configuration for rip.
        Use base `add_to_config` directly for more complex configurations
        """
        self.add_rip()
        self.add_version(2)
        for interface in self.interfaces:
            self.add_network(interface.id)

        self.create_config()

    def run(self):
        """
        Runs the ripd command
        """
        run_ripd(self.router_ns_id, self.conf_file, self.pid_file)



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Plot ping results"""

import matplotlib.pyplot as plt

from ..pack import Pack
from .common import simple_plot


def _plot_ping_flow(flow, node, dest):
    """
    Plot ping stats of the flow

    Parameters
    ----------
    exp_name : str
        Name of experiment for which results were obtained
    flow : List
        List with timestamps and stats
    node : str
        Node from which ping results were obtained from
    dest :
        Destination ip:port address of the flow

    Returns
    -------
    tuple
        Timestamped rtt values
    """
    # "meta" item will always be present, hence `<= 1`
    if len(flow) <= 1:
        raise ValueError('Flow from {} to destination {}'
                         'doesn\'t have any parsed ping result.'.format(node,
                                                                        dest))

    # First item is the "meta" item with user given information
    user_given_start_time = float(flow[0]['start_time'])

    # "Bias" actual start_time in experiment with user given start time
    start_time = float(flow[1]['timestamp']) - user_given_start_time

    timestamp = []
    rtt = []

    for data in flow[1:]:
        rtt.append(float(data['rtt']))
        relative_time = float(data['timestamp']) - start_time
        timestamp.append(relative_time)

    fig = simple_plot('Ping', timestamp, rtt, 'Time (s)', 'RTT (ms)',
                      legend_string=f'{node} to {dest}')
    filename = '{node}_{dest}_ping.png'.format(node=node, dest=dest)
    Pack.dump_plot('ping', filename, fig)
    plt.close(fig)

    return (timestamp, rtt)


def plot_ping(parsed_data):
    """
    Plot statistics obtained from ping

    Parameters
    ----------
    exp_name : str
        Name of experiment for which results were obtained
    parsed_data : Dict
        JSON data parsed from ss
    """
    for node in parsed_data:
        node_data = parsed_data[node]

        for connection in node_data:
            for dest in connection:
                flow = connection[dest]
                _plot_ping_flow(flow, node, dest)



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Handles addresses of interfaces"""

import ipaddress

class Address:
    """Validate address.

    TODO: Untested for IPv6
    """

    def __init__(self, addr_str):
        """
        The constructor validates the entered ip address.

        Parameters
        ----------
        addr_str : str
            The string representing the address
        """

        if addr_str.isupper():      # A special string in place of address
            if addr_str == 'DEFAULT':
                self.ip_addr = addr_str.lower()
            else:
                raise ValueError(
                    addr_str+' is not a special IP string. Perhaps you meant DEFAULT?')
        elif '/' in addr_str:       # An interface address
            # raises exception in invalid case
            ipaddress.ip_interface(addr_str)
            self.ip_addr = addr_str
        else:                       # An IP address
            # raises exception in invalid case
            ipaddress.ip_address(addr_str)
            self.ip_addr = addr_str+'/32'

    def get_addr(self, with_subnet=True):
        """Getter for ip_addr

        Parameters
        ----------
        with_subnet :
             (Default value = True)

        Returns
        -------
        str
            IP address as string
        """
        if with_subnet:
            return self.ip_addr

        return self.ip_addr.split('/')[0]

    def get_subnet(self):
        """Get the subnet of the given address"""
        if self.ip_addr == 'default':
            raise Exception("default address cannot have a subnet")
        interface = ipaddress.ip_interface(self.ip_addr)
        return interface.network.compressed

    def is_subnet(self):
        """Check if the address is a subnet or not"""
        if self.ip_addr == 'default':
            return False
        if self.ip_addr == self.get_subnet():
            return True
        return False

    def __repr__(self):
        classname = self.__class__.__name__
        return f'{classname}({self.ip_addr!r})'


class Subnet:
    """For generating sequential addresses in a subnet

    NOTE: Supported only for IPv4 addresses
    """

    def __init__(self, addr_str):
        address = Address(addr_str)

        if address.is_subnet():
            self._net_addr = ipaddress.ip_network(addr_str)
        else:
            raise ValueError('Parameter is not a subnet/network address')

        self._counter = 0

    def get_next_addr(self):
        """Get next address in sequence in the subnet"""
        self._counter += 1
        try:
            address = self._net_addr[self._counter]
        except IndexError as err:
            raise ValueError('All the addresses for the network have been exhausted') from err
        return Address(address.compressed + '/' + str(self._net_addr.prefixlen))

    @property
    def counter(self):
        """Getter for counter"""
        return self._counter

    @property
    def network_address(self):
        """Getter for network address"""
        return self._net_addr

    def __repr__(self):
        classname = self.__class__.__name__
        return f'{classname}({self._net_addr!r})'



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Plot netperf results"""

import matplotlib.pyplot as plt

from ..pack import Pack
from .common import simple_plot, mix_plot


def _plot_netperf_flow(flow, node, dest):
    """
    Plot netperf stats of the flow

    Parameters
    ----------
    exp_name : str
        Name of experiment for which results were obtained
    flow : List
        List with timestamps and stats
    node : str
        Node from which netperf results were obtained from
    dest :
        Destination ip:port address of the flow

    Returns
    -------
    tuple
        Timestamped sending_rate values
    """
    # "meta" item will always be present, hence `<= 1`
    if len(flow) <= 1:
        raise ValueError('Flow from {} to destination {}'
                         'doesn\'t have any parsed ss result.'.format(node,
                                                                      dest))

    # First item is the "meta" item with user given information
    user_given_start_time = float(flow[0]['start_time'])

    # "Bias" actual start_time in experiment with user given start time
    start_time = float(flow[1]['timestamp']) - user_given_start_time

    timestamp = []
    sending_rate = []

    for data in flow[1:]:
        sending_rate.append(float(data['sending_rate']))
        relative_time = float(data['timestamp']) - start_time
        timestamp.append(relative_time)

    # TODO: Check if sending_rate is always in Mbps
    fig = simple_plot('Netperf', timestamp, sending_rate, 'Time (s)', 'Sending rate (Mbps)',
                      legend_string=f'{node} to {dest}')

    filename = '{node}_{dest}_sending_rate.png'.format(node=node, dest=dest)
    Pack.dump_plot('netperf', filename, fig)
    plt.close(fig)

    return (timestamp, sending_rate)


def plot_netperf(parsed_data):
    """
    Plot statistics obtained from netperf

    Parameters
    ----------
    exp_name : str
        Name of experiment for which results were obtained
    parsed_data : Dict
        JSON data parsed from ss
    """
    for node in parsed_data:
        node_data = parsed_data[node]

        all_flow_data = []
        for connection in node_data:
            for dest in connection:
                flow = connection[dest]
                values = _plot_netperf_flow(flow, node, dest)
                all_flow_data.append({'label': f'{node} to {dest}', 'values': values})

        fig = mix_plot('Netperf', all_flow_data, 'Time (s)',
                       'Sending rate (Mbps)', with_sum=True)
        filename = f'{node}_sending_rate.png'
        Pack.dump_plot('netperf', filename, fig)
        plt.close(fig)



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Handle plotting results obtained"""

import matplotlib.style as style

# Set plot style
style.use('seaborn-paper')
style.use('ggplot')



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""User API to setup and run experiments on a given topology"""

import copy
import logging
from ..topology import Address
from .run_exp import run_experiment
from .pack import Pack

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

#pylint: disable=too-many-instance-attributes
#pylint: disable=too-few-public-methods
class Flow():
    """Defines a flow in the topology"""

    #pylint: disable=too-many-arguments
    def __init__(self, source_node, destination_node, destination_address, start_time, stop_time,
                 number_of_flows):
        """
        'Flow' object in the topology

        Parameters
        ----------
        source_node : Node
            Source node of flow
        destination_node : Node
            Destination node of flow
        destination_address : Address/str
            Destination address of flow
        start_time : int
            Time to start flow (in seconds)
        stop_time : int
            Time to stop flow (in seconds)
        number_of_flows : int
            Number of flows
        """
        self.source_node = source_node
        self.destination_node = destination_node
        self.destination_address = destination_address
        self.start_time = start_time
        self.stop_time = stop_time
        self.number_of_flows = number_of_flows

        self._options = {
            'protocol': 'TCP',
            'cong_algo': 'cubic'
        }

    @property
    def destination_address(self):
        """Getter for destination address"""
        return self._destination_address

    @destination_address.setter
    def destination_address(self, destination_address):
        """Setter for destination address"""
        if isinstance(destination_address, str):
            destination_address = Address(destination_address)
        self._destination_address = destination_address

    def _get_props(self):
        """
        Get flow properties.

        NOTE: To be used internally
        """

        return [self.source_node.id, self.destination_node.id,
                self.destination_address.get_addr(with_subnet=False),
                self.start_time, self.stop_time, self.number_of_flows, self._options]

    def __repr__(self):
        classname = self.__class__.__name__
        return (f'{classname}({self.source_node!r}, {self.destination_node!r},'
                f' {self.destination_address!r}), {self.start_time!r}, {self.stop_time!r}'
                f' {self.number_of_flows!r})')

class Experiment():
    """Handles experiment to be run on topology"""

    def __init__(self, name):
        """
        Create experiment

        Parameters
        ----------
        name : str
            Name of experiment
        """
        self.name = name
        self.flows = []
        self.node_stats = []
        self.qdisc_stats = []

    def add_flow(self, flow):
        """
        Add flow to experiment
        By default, the flow is assumed to be
        TCP with cubic congestion algorithm

        Parameters
        ----------
        flow : Flow
            Add flow to experiment
        """
        self.flows.append(copy.deepcopy(flow))

    def add_tcp_flow(self, flow, congestion_algorithm='cubic'):
        """
        Add TCP flow to experiment

        Parameters
        ----------
        flow : Flow
            Flow to be added to experiment
        congestion_algorithm : str
            TCP congestion algorithm (Default value = 'cubic')
        """

        # TODO: Verify congestion algorithm

        options = {
            'protocol': 'TCP',
            'cong_algo': congestion_algorithm
        }

        flow._options = options #pylint: disable=protected-access
        self.add_flow(flow)

    def add_udp_flow(self, flow, target_bandwidth='1mbit'):
        """
        Add UDP flow to experiment

        Parameters
        ----------
        flow : Flow
            Flow to be added to experiment
        target_bw :
            UPD bandwidth (in Mbits) (Default value = '1mbit')
        """
        if len(target_bandwidth) < 4 or target_bandwidth[-4:] != 'mbit':
            raise Exception('Invalid bandwidth unit given to target_bandwidth parameter. ' \
                            'Expecting mbit.')

        options = {
            'protocol': 'UDP',
            'target_bw': target_bandwidth
        }

        flow._options = options #pylint: disable=protected-access
        self.add_flow(flow)

    def require_qdisc_stats(self, interface, stats=''):
        """
        Stats to be obtained from qdisc in interface

        Parameters
        ----------
        interface : Interface
            Interface containing the qdisc
        stats : list(str)
            Stats required (Default value = '') [NOT SUPPORTED]
        """
        # TODO: Leads to rewrite if the function is called
        # twice with same 'interface'

        # for stat in stats:
        #     if stat not in Experiment.qdisc_stats:
        #         raise ValueError('{} is not a valid Queue property.'.format(stat))

        if interface.get_qdisc() is None:
            raise ValueError(
                'Given interface hasn\'t been assigned any qdisc.')

        self.qdisc_stats.append({
            'ns_id': interface.node.id,
            'int_id': interface.ifb.id,
            'qdisc': interface.get_qdisc().qdisc,
            'stats': stats
        })

    def run(self):
        """Run the experiment"""
        logger.info('Running experiment %s \n', self.name)
        Pack.init(self.name)
        run_experiment(self)

    def __repr__(self):
        classname = self.__class__.__name__
        return f'{classname}({self.name!r})'



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Handle traffic control entities"""

from .. import engine

# TODO: Improve this module such that the below pylint disables are no
# longer required

#pylint: disable=too-few-public-methods
#pylint: disable=too-many-arguments
#pylint: disable=too-many-instance-attributes

class Qdisc:
    """Handle Queueing Discipline"""

    def __init__(self, namespace_id, dev_id, qdisc, parent='root', handle='', **kwargs):
        """
        Constructor to add a qdisc (Queueing Discipline) to an interface (device)

        Parameters
        ----------
        namespace_id : str
            The id of the namespace to which the interface belongs to
        dev_id : str
            The id of the interface to which the qdisc is to be added
        qdisc : str
            The qdisc which needs to be added to the interface
        dev_id : str
            The id of the interface to which the qdisc is to be added
        namespace_id : str
            The id of the namespace that the device belongs to
        parent : str
            id of the parent class in major:minor form(optional)
        handle : str
            id of the filter
        """
        self.namespace_id = namespace_id
        self.dev_id = dev_id
        self.qdisc = qdisc
        self.parent = parent
        self.handle = handle

        engine.add_qdisc(namespace_id, dev_id, qdisc, parent, handle, **kwargs)


class Class:
    """Handle classes associated to qdisc"""

    def __init__(self, namespace_id, dev_id, qdisc, parent='root', classid='', **kwargs):
        """
        Constructor to create an object that represents a class

        Parameters
        ----------
        namespace_id : str
            The id of the namespace to which the interface belongs to
        dev_id : str
            The id of the interface to which the qdisc is to be added
        namespace_id : str
            The id of the namespace that the device belongs to
        qdisc : str
            The qdisc which needs to be added to the interface
        parent : str
            id of the parent class in major:minor form(optional)
        classid : str
            id of the class
        """
        self.namespace_id = namespace_id
        self.dev_id = dev_id
        self.qdisc = qdisc          # NOTE: should be renamed to knid
        self.parent = parent
        self.classid = classid

        engine.add_class(namespace_id, dev_id, parent,
                         qdisc, classid, **kwargs)


class Filter:
    """Handle filters to assign to class/qdisc"""

    def __init__(self, namespace_id, dev_id, protocol, priority, filtertype, flowid,
                 parent='root', handle='', **kwargs):
        """
        Constructor to design a Filter to assign to a Class or Qdisc

        Parameters
        ----------
        namespace_id : str
            The id of the namespace to which the interface belongs to
        dev_id : str
            The id of the interface to which the qdisc is to be added
        protocol : str
            protocol used
        priority : int
            priority of the filter
        filtertype : str
            one of the available filters
        flowid : Class
            classid of the class where the traffic is enqueued
            if the traffic passes the filter
        parent : str
            id of the parent class in major:minor form(optional)
        handle : str
            id of the filter
        filter : dictionary
            filter parameters
        """
        self.namespace_id = namespace_id
        self.dev_id = dev_id
        self.protocol = protocol
        self.priority = priority
        self.filtertype = filtertype
        self.flowid = flowid
        self.parent = parent
        self.handle = handle

        engine.add_filter(namespace_id, dev_id, protocol, priority,
                          filtertype, parent, handle, **kwargs)



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Test topology map API"""

import unittest
from nest.topology_map import TopologyMap

#pylint: disable=missing-docstring
#pylint: disable=too-many-instance-attributes

class TestTopologyMap(unittest.TestCase):

    def setUp(self):
        # Define namespace id and names
        self.ns_id1 = 'ns_id1'
        self.ns_id2 = 'ns_id2'
        self.ns_name1 = 'ns_name1'
        self.ns_name2 = 'ns_name2'

        # Add namespaces
        TopologyMap.add_namespace(self.ns_id1, self.ns_name1)
        TopologyMap.add_namespace(self.ns_id2, self.ns_name2)

        # Define interface id and names
        self.int_id1 = 'int_id1'
        self.int_id2 = 'int_id2'
        self.int_name1 = 'int_name1'
        self.int_name2 = 'int_name2'

        # Add interfaces
        TopologyMap.add_interface(self.ns_id1, self.int_id1, self.int_name1)
        TopologyMap.add_interface(self.ns_id1, self.int_id2, self.int_name2)

    def tearDown(self):
        TopologyMap.delete_all_mapping()

    def test_add_and_get_namespace(self):
        self.assertEqual(TopologyMap.get_namespace(self.ns_id1)['name'], self.ns_name1)
        self.assertEqual(TopologyMap.get_namespace(self.ns_id2)['name'], self.ns_name2)

    def test_add_and_get_interface(self):
        self.assertEqual(TopologyMap.get_interface(self.ns_id1, self.int_id1)['name'],
                         self.int_name1)
        self.assertEqual(TopologyMap.get_interface(self.ns_id1, self.int_id2)['name'],
                         self.int_name2)

    def test_add_same_entity_again(self):
        with self.assertRaises(ValueError):
            TopologyMap.add_namespace(self.ns_id1, self.ns_name1)
        with self.assertRaises(ValueError):
            TopologyMap.add_interface(self.ns_id1, self.int_id1, self.int_name1)

if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Class to handle zebra"""

from nest.routing.quagga_base import QuaggaBase
from nest.engine.quagga import run_zebra


class Zebra(QuaggaBase):
    """
    Handles zebra related functionalities for quagga.
    Refer to `QuaggaBase` for usage
    """

    def __init__(self, router_ns_id, interfaces, conf_dir):
        super().__init__(router_ns_id, interfaces, 'zebra', conf_dir)

    def add_interface(self, interface):
        """
        Add interface command to config file
        """
        self.add_to_config(f'interface {interface}')

    def add_ip_address(self, ip_address):
        """
        Add ip address command to config file
        """
        self.add_to_config(f' ip address {ip_address}')

    def create_basic_config(self):
        """
        Creates a file with basic configuration for ospf.
        Use base `add_to_config` directly for more complex configurations
        """
        for interface in self.interfaces:
            self.add_interface(interface.id)
            self.add_ip_address(interface.address.get_addr())
        self.create_config()

    def run(self):
        """
        Runs the zebra daemon
        """
        run_zebra(self.router_ns_id, self.conf_file, self.pid_file)



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Execute linux commands"""

import logging
import shlex
import subprocess

LOG_COMMANDS = False # TODO: Should be config parameter

logger = logging.getLogger(__name__)
if LOG_COMMANDS:
    # pylint: disable=no-member
    logger.setLevel(logging.TRACE)
    fh = logging.FileHandler('commands.sh', 'w')
    fh.setLevel(logging.TRACE)
    formatter = logging.Formatter('%(message)s')
    fh.setFormatter(formatter)
    logger.addHandler(fh)

# pylint: disable=inconsistent-return-statements
def exec_subprocess(cmd, shell=False, output=False):
    """
    Executes a command

    Parameters
    ----------
    cmd : str
        command to be executed
    shell : boolean
        Spawns a shell and executes the command if true
        (Default value = False)
    output : boolean
        True if the output of the `cmd` is to be returned
        (Default value = False)

    Returns
    -------
    int
        Return code recieved after executing the command
    """

    # TODO: Commands with pipes are easily executed when
    # they are run within a shell.
    # But it may not be the most efficient way.

    temp_cmd = cmd
    if shell is False:
        temp_cmd = cmd.split()
    proc = subprocess.Popen(temp_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=shell)

    (stdout, _) = proc.communicate()
    logger.trace(cmd)

    if output:
        return stdout.decode()
    return proc.returncode

def exec_exp_commands(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, timeout=None):
    """
    executes experiment related commands like ss, tc and netperf

    Parameters
    ----------
    cmd : str
        command to be executed
    stdout : File
        temp file(usually) to store the output (Default value = subprocess.PIPE)
    stderr : File
        temp file(usually) to store errors, if any (Default value = subprocess.PIPE)
    timeout :
         (Default value = None)

    Returns
    -------


    """
    proc = subprocess.Popen(shlex.split(cmd), stdout=stdout, stderr=stderr)
    try:
        proc.communicate(timeout=timeout)
    except subprocess.TimeoutExpired:
        proc.kill()
        stderr.write(b'Connection timeout')
    return proc.returncode



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""
Base class for other runners
"""

import tempfile
import logging
from ...engine import exec_exp_commands


class Runner:
    """
    Base class for other runners

    Attributes
    ----------
    out : File
        temporary file to hold the stats
    err : File
        temporary file to hold any errors
    start_time : num
            time at which netperf is to run
    run_time : num
        total time to run netperf for
    """

    def __init__(self, start_time, run_time):
        self.out = tempfile.TemporaryFile()
        self.err = tempfile.TemporaryFile()

        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(logging.INFO)

        self.start_time = start_time
        self.run_time = run_time

    def run(self, command):
        """
        Runs the `command` and stores stdout or stderr
        """
        return_code = exec_exp_commands(
            command, stdout=self.out, stderr=self.err)
        if return_code != 0:
            self.print_error()

    def print_error(self):
        """
        Method to print error from `self.err`.
        Should be overriden by base class
        """
        self.logger.error('Unknown error occured')

    def get_meta_item(self):
        """
        Return the meta item for the given flow.
        This "meta" information is required by plotter.
        """
        meta_item = {
            'meta': True,
            'start_time': str(self.start_time),
            'stop_time': str(self.start_time + self.run_time)
        }
        return meta_item

    def clean_up(self):
        """
        Closes the temp files created
        """
        self.out.close()
        self.err.close()



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

# Get version
version = {}
with open("nest/version.py") as fp:
    exec(fp.read(), version)

# Get the long description from the README file
long_description = (here / 'README.md').read_text(encoding='utf-8')

setup(
    name='nitk-nest',
    version=version['__version__'],
    description='NeST: Network Stack Tester',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://nitk-nest.github.io/',
    author='NITK',
    author_email='nest.nitk@gmail.com',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Intended Audience :: System Administrators',
        'Topic :: Education',
        'Topic :: Scientific/Engineering',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    keywords='network, namespace, linux',
    packages=find_packages(),
    package_data={
        "nest.experiment.parser": ["iterators/*.sh"],
        "nest": ["config.json"]
    },
    python_requires='>=3.6, <4',
    install_requires=[
        'matplotlib',
        'numpy',
        'packaging'
    ],
    project_urls={
        'Source': 'https://gitlab.com/nitk-nest/nest',
    },
)


# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Plot ss results"""

import matplotlib.pyplot as plt

from ..pack import Pack
from .common import simple_plot, mix_plot


def _get_list_of_ss_params():
    """
    Return list of params parsed by ss
    """
    return ['cwnd', 'rtt', 'dev_rtt', 'ssthresh', 'rto', 'delivery_rate', 'pacing_rate']

def _get_ss_params_units():
    """
    Obtain the units of each ss parameters
    """
    units = {
        'cwnd': None,
        'rtt': 'ms',
        'dev_rtt': 'ms',
        'ssthresh': None,
        'rto': 'ms',
        'delivery_rate': 'Mbps',
        'pacing_rate': 'Mbps'
    }
    return units

def _get_nicer_param_names():
    """
    For example,
    'rtt' -> 'RTT'
    'delivery_rate' -> 'Delivery rate'
    """
    nice_names = {
        'cwnd': 'CWND',
        'rtt': 'RTT',
        'dev_rtt': 'DevRTT',
        'ssthresh': 'ssthresh',
        'rto': 'RTO',
        'delivery_rate': 'Delivery rate',
        'pacing_rate': 'Pacing rate'
    }
    return nice_names

def _get_ylabel(param):
    """
    Obtain y label for plots for a given parameter.

    For eg. 'delivery_rate' -> 'Delivery rate (Mbps)'

    Arguments
    ---------
        param : str
            parameter

    Returns
    -------
        str
    """
    # Get ss param units
    units = _get_ss_params_units()
    # Get "nicer" param names
    nicer_param_names = _get_nicer_param_names()

    ylabel = nicer_param_names[param]
    if units[param] is not None:
        ylabel += f' ({units[param]})'

    return ylabel


def _extract_from_ss_flow(flow, node, dest_ip, dest_port):
    """
    Extract information from flow data and convert it to
    conviniently plottable data format

    Parameters
    ----------
    flow : List
        List with timestamps and stats
    node : string
        Node from which ss results were obtained from
    dest_ip : string
        Destination ip address of the flow
    dest_port : string
        Destination port of the flow

    Returns
    -------
    (timestamp, flow_params)
        Return timestamp and flow parameters
    """
    # "meta" item will always be present, hence `<= 1`
    if len(flow) <= 1:
        raise ValueError('Flow from {} to destination {}:{}'
                         'doesn\'t have any parsed ss result.'.format(node,
                                                                      dest_ip, dest_port))

    # First item is the "meta" item with user given information
    user_given_start_time = float(flow[0]['start_time'])

    # "Bias" actual start_time in experiment with user given start time
    start_time = float(flow[1]['timestamp']) - user_given_start_time

    timestamp = []
    flow_params = {}

    for param in _get_list_of_ss_params():
        flow_params[param] = []

    for data in flow[1:]:
        for stat in flow_params:
            if stat in data:
                flow_params[stat].append(float(data[stat]))
            else:
                flow_params[stat].append(None)
        relative_time = float(data['timestamp']) - start_time
        timestamp.append(relative_time)

    return (timestamp, flow_params)


def _plot_ss_flow(flow, node, dest_ip, dest_port):
    """
    Plot ss stats of the flow

    Parameters
    ----------
    exp_name : string
        Name of experiment for which results were obtained
    flow : List
        List with timestamps and stats
    node : string
        Node from which ss results were obtained from
    dest_ip : string
        Destination ip address of the flow
    dest_port : string
        Destination port of the flow
    """
    (timestamp, flow_params) = _extract_from_ss_flow(
        flow, node, dest_ip, dest_port)

    for param in flow_params:
        fig = simple_plot('Socket Stats (ss)', timestamp, flow_params[param], 'Time (s)',
                          _get_ylabel(param), legend_string=f'{node} to {dest_ip}:{dest_port}')

        filename = f'{node}_{dest_ip}:{dest_port}_{param}.png'
        Pack.dump_plot('ss', filename, fig)
        plt.close(fig)

    return (timestamp, flow_params)

# pylint: disable=too-many-locals
def plot_ss(parsed_data):
    """
    Plot statistics obtained from ss

    Parameters
    ----------
    exp_name : string
        Name of experiment for which results were obtained
    parsed_data : Dict
        JSON data parsed from ss
    """

    # pylint: disable=too-many-nested-blocks
    for node in parsed_data:
        node_data = parsed_data[node]
        for connection in node_data:
            for dest_ip in connection:

                all_flow_data = []

                flow_data = connection[dest_ip]
                for dest_port in flow_data:
                    flow = flow_data[dest_port]
                    values = _plot_ss_flow(flow, node, dest_ip, dest_port)
                    all_flow_data.append({
                        'values': values,
                        'label': f'{node} to {dest_ip}:{dest_port}'})

                if len(all_flow_data) > 0:

                    x_vals = []
                    labels = []

                    for flow_data in all_flow_data:
                        x_vals.append(flow_data['values'][0])
                        labels.append(flow_data['label'])

                    for param in all_flow_data[0]['values'][1]:
                        y_vals = []
                        for flow_data in all_flow_data:
                            y_vals.append(flow_data['values'][1][param])

                        data = []

                        for i in range(len(labels)):
                            data.append(
                                {'values': (x_vals[i], y_vals[i]), 'label': labels[i]})

                        fig = mix_plot('Socket Stats (ss)', data, 'Time (s)', _get_ylabel(param))
                        filename = f'{node}_{dest_ip}_{param}.png'
                        Pack.dump_plot('ss', filename, fig)
                        plt.close(fig)



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Script to be run for running experiments on topology"""

from multiprocessing import Process
from collections import namedtuple, defaultdict
import logging

from nest import config
from ..topology_map import TopologyMap
from ..clean_up import kill_processes
# Import results
from .results import SsResults, NetperfResults, TcResults, PingResults
# Import parsers
from .parser.ss import SsRunner
from .parser.netperf import NetperfRunner
from .parser.tc import TcRunner
from .parser.ping import PingRunner
# Import plotters
from .plotter.ss import plot_ss
from .plotter.netperf import plot_netperf
from .plotter.tc import plot_tc
from .plotter.ping import plot_ping
from ..experiment.parser.iperf import IperfRunner
from ..engine.util import is_dependency_installed

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

#pylint: disable=too-many-locals


def run_experiment(exp):
    """
    Run experiment

    Parameters
    -----------
    exp : Experiment
        The experiment attributes
    """

    # Could be moved to config?
    tools = ['netperf', 'ss', 'tc', 'iperf3', 'ping']
    exp_workers = []    # Processes to setup flows and statistics collection
    Runners = namedtuple('runners', tools)
    exp_runners = Runners(netperf=[], ss=[], tc=[],
                          iperf3=[], ping=[])  # Runner objects

    # Keep track of all destination nodes [to ensure netperf and iperf
    # server is run atmost once]
    destination_nodes = {
        'netperf': set(),
        'iperf3' : set()
    }

    # Contains start time and end time to run respective command
    # from a source netns to destination addr
    ss_schedules = defaultdict(lambda: (float('inf'), float('-inf')))
    ping_schedules = defaultdict(lambda: (float('inf'), float('-inf')))

    # exp_start = float('inf')
    exp_end_t = float('-inf')

    dependencies = get_dependency_status(tools)

    # Traffic generation
    for flow in exp.flows:
        # Get flow attributes
        [src_ns, dst_ns, dst_addr, start_t,
         stop_t, _, options] = flow._get_props()  # pylint: disable=protected-access

        # exp_start = min(exp_start, start_t)
        exp_end_t = max(exp_end_t, stop_t)

        (min_start, max_stop) = ping_schedules[(src_ns, dst_addr)]
        ping_schedules[(src_ns, dst_addr)] = (
            min(min_start, start_t), max(max_stop, stop_t))

        # Setup TCP/UDP flows
        if options['protocol'] == 'TCP':
            dependencies['netperf'], tcp_runners, tcp_workers, ss_schedules = setup_tcp_flows(
                dependencies['netperf'], flow, ss_schedules, destination_nodes['netperf'])

            exp_runners.netperf.extend(tcp_runners)
            exp_workers.extend(tcp_workers)

            # Update destination nodes
            destination_nodes['netperf'].add(dst_ns)

        elif options['protocol'] == 'UDP':
            dependencies['iperf3'], udp_runners, upd_workers = setup_udp_flows(
                dependencies['iperf3'], flow, ss_schedules, destination_nodes['iperf3'])

            exp_runners.iperf3.extend(udp_runners)
            exp_workers.extend(upd_workers)

            # Update destination nodes
            destination_nodes['iperf3'].add(dst_ns)

    if dependencies['netperf'] == 1:
        ss_workers, ss_runners = setup_ss_runners(dependencies['ss'],
                                                  ss_schedules)
        exp_workers.extend(ss_workers)
        exp_runners.ss.extend(ss_runners)

        tc_workers, tc_runners = setup_tc_runners(
            dependencies['tc'], exp.qdisc_stats, exp_end_t)
        exp_workers.extend(tc_workers)
        exp_runners.ss.extend(tc_runners)

    ping_workers, ping_runners = setup_ping_runners(
        dependencies['ping'], ping_schedules)
    exp_workers.extend(ping_workers)
    exp_runners.ping.extend(ping_runners)

    # Start traffic generation and parsing
    run_workers(exp_workers)

    logger.info('Experiment complete!')
    logger.info("Parsing statistics...")

    # Parse the stored statistics
    run_workers(get_parser_workers(exp_runners))

    logger.info('Output results as JSON dump')

    # Output results as JSON dumps
    dump_json_ouputs()

    if config.get_value('plot_results'):
        logger.info('Plotting results...')

        # Plot results and dump them as images
        run_workers(get_plotter_workers())

        logger.info('Plotting complete!')

    cleanup()


def run_workers(workers):
    """
    Run and wait for processes to finish

    Parameters
    ----------
    workers: list[multiprocessing.Process]
        List of processes to be run
    """
    # Start workers
    for worker in workers:
        worker.start()

    # wait for all the workers to finish
    for worker in workers:
        worker.join()


def get_plotter_workers():
    """
    Setup plotting processes

    Returns
    -------
    List[multiprocessing.Process]
        plotters
    """
    plotters = []

    plotters.append(Process(target=plot_ss, args=(SsResults.get_results(),)))
    plotters.append(Process(target=plot_netperf,
                            args=(NetperfResults.get_results(),)))
    plotters.append(Process(target=plot_tc, args=(TcResults.get_results(),)))
    plotters.append(
        Process(target=plot_ping, args=(PingResults.get_results(),)))

    return plotters


def dump_json_ouputs():
    """
    Outputs experiment results as json dumps
    """
    SsResults.output_to_file()
    NetperfResults.output_to_file()
    TcResults.output_to_file()
    PingResults.output_to_file()


def get_parser_workers(runners_list):
    """
    Setup parsing processes

    Parameters
    ----------
    runners_list: collections.NamedTuple
        all(netperf, ping, ss, tc..) the runners

    Returns
    -------
    List[multiprocessing.Process]
        parsers
    """
    runners = []

    for ss_runner in runners_list.ss:
        runners.append(Process(target=ss_runner.parse))

    for netperf_runner in runners_list.netperf:
        runners.append(Process(target=netperf_runner.parse))

    for tc_runner in runners_list.tc:
        runners.append(Process(target=tc_runner.parse))

    for ping_runner in runners_list.ping:
        runners.append(Process(target=ping_runner.parse))

    return runners


def get_dependency_status(tools):
    """
    Checks for dependency

    Parameters
    ----------
    tools: List[str]
        list of tools to check for it's installation

    Returns
    -------
    dict
        contains information as to whether `tools` are installed
    """
    dependencies = {}
    for dependency in tools:
        dependencies[dependency] = int(is_dependency_installed(dependency))
    return dependencies


def setup_tcp_flows(dependency, flow, ss_schedules, destination_nodes):
    """
    Setup netperf to run tcp flows
    Parameters
    ----------
    dependency: int
        whether netperf is installed
    flow: Flow
        Flow parameters
    ss_schedules:
        ss_schedules so far
    destination_nodes:
        Destination nodes so far already running netperf server

    Returns
    -------
    dependency: int
        updated dependency incase netperf is not installed
    netperf_runners: List[NetperfRunner]
        all the netperf flows generated
    workers: List[multiprocessing.Process]
        Processes to run netperf flows
    ss_schedules: dict
        updated ss_schedules
    """
    netperf_runners = []
    workers = []
    if dependency == 0:
        logger.warning('Netperf not found. Tcp flows cannot be generated')
        # To avoid duplicate warning messages
        dependency = 2
    elif dependency == 1:
        # Get flow attributes
        [src_ns, dst_ns, dst_addr, start_t, stop_t,
         n_flows, options] = flow._get_props()  # pylint: disable=protected-access

        # Run netserver if not already run before on given dst_node
        if dst_ns not in destination_nodes:
            NetperfRunner.run_netserver(dst_ns)

        src_name = TopologyMap.get_namespace(src_ns)['name']

        netperf_options = {}
        netperf_options['testname'] = 'TCP_STREAM'
        netperf_options['cong_algo'] = options['cong_algo']

        logger.info('Running %s netperf flows from %s to %s...',
                    n_flows, src_name, dst_addr)

        # Create new processes to be run simultaneously
        for _ in range(n_flows):
            runner_obj = NetperfRunner(
                src_ns, dst_addr, start_t, stop_t-start_t, **netperf_options)
            netperf_runners.append(runner_obj)
            workers.append(Process(target=runner_obj.run))

        # Find the start time and stop time to run ss command in `src_ns` to a `dst_addr`
        ss_schedules = _find_start_stop_time_for_ss(src_ns, dst_addr, start_t, stop_t, ss_schedules)

    return dependency, netperf_runners, workers, ss_schedules


def setup_udp_flows(dependency, flow, ss_schedules, destination_nodes):
    """
    Setup iperf3 to run udp flows

    Parameters
    ----------
    dependency: int
        whether iperf3 is installed
    flow: Flow
        Flow parameters
    ss_schedules:
        ss_schedules so far
    destination_nodes:
        Destination nodes so far already running iperf3 server

    Returns
    -------
    depedency: int
        updated dependency incase iproute2 is not installed
    iperf3_runners: List[NetperfRunner]
        all the iperf3 udp flows generated
    workers: List[multiprocessing.Process]
        Processes to run iperf3 udp flows
    """
    iperf3_runners = []
    workers = []
    if dependency == 0:
        logger.warning(
            'Iperf3 not found. Udp flows cannot be generated')
        # To avoid duplicate warning messages
        dependency = 2
    elif dependency == 1:
        # Get flow attributes
        [src_ns, dst_ns, dst_addr, start_t, stop_t,
         n_flows, options] = flow._get_props()  # pylint: disable=protected-access

        # Run iperf3 server if not already run before on given dst_node
        if dst_ns not in destination_nodes:
            IperfRunner.run_server(dst_ns)

        src_name = TopologyMap.get_namespace(src_ns)['name']

        logger.info('Running %s udp flows from %s to %s...',
                    n_flows, src_name, dst_addr)

        runner_obj = IperfRunner(src_ns, dst_addr, options['target_bw'], n_flows,
                                 start_t, stop_t-start_t)
        iperf3_runners.append(runner_obj)
        workers.append(Process(target=runner_obj.run))

        # Find the start time and stop time to run ss command in `src_ns` to a `dst_addr`
        ss_schedules = _find_start_stop_time_for_ss(src_ns, dst_addr, start_t, stop_t, ss_schedules)

    return dependency, iperf3_runners, workers


def setup_ss_runners(dependency, ss_schedules):
    """
    setup SsRunners for collecting tcp socket statistics

    Parameters
    ----------
    dependency: int
        whether ss is installed
    ss_schedules: dict
        start time and end time for SsRunners

    Returns
    -------
    workers: List[multiprocessing.Process]
        Processes to run ss at nodes
    runners: List[SsRunners]
    """
    runners = []
    workers = []
    if dependency == 1:
        logger.info('Running ss on nodes...')
        for ns_id, timings in ss_schedules.items():
            ss_runner = SsRunner(ns_id[0], ns_id[1], timings[0],
                                 timings[1] - timings[0])
            runners.append(ss_runner)
            workers.append(Process(target=ss_runner.run))
    else:
        logger.warning(
            'ss not found. Sockets stats will not be collected')
    return workers, runners


def setup_tc_runners(dependency, qdisc_stats, exp_end):
    """
    setup TcRunners for collecting qdisc statistics

    Parameters
    ----------
    dependency: int
        whether tc is installed
    qdisc_stats: dict
        info regarding nodes to run tc on
    exp_end: float
        time to stop running tc
    Returns
    -------
    workers: List[multiprocessing.Process]
        Processes to run tc at nodes
    runners: List[TcRunners]
    """
    runners = []
    workers = []
    if dependency == 1 and len(qdisc_stats) > 0:
        logger.info('Running tc on requested interfaces...')
        for qdisc_stat in qdisc_stats:
            tc_runner = TcRunner(qdisc_stat['ns_id'], qdisc_stat['int_id'],
                                 qdisc_stat['qdisc'], exp_end)
            runners.append(tc_runner)
            workers.append(Process(target=tc_runner.run))
    elif dependency != 1:
        logger.warning('tc not found. Qdisc stats will not be collected')
    return workers, runners


def setup_ping_runners(dependency, ping_schedules):
    """
    setup PingRunners for collecting latency

    Parameters
    ----------
    dependency: int
        whether ping is installed
    ping_schedules: dict
        start time and end time for PingRunners

    Returns
    -------
    workers: List[multiprocessing.Process]
        Processes to run ss at nodes
    runners: List[SsRunners]
    """
    runners = []
    workers = []
    if dependency == 1:
        for ns_id, timings in ping_schedules.items():
            ping_runner = PingRunner(
                ns_id[0], ns_id[1], timings[0], timings[1]-timings[0])
            runners.append(ping_runner)
            workers.append(Process(target=ping_runner.run))
    else:
        logger.warning('ping not found')
    return workers, runners


def cleanup():
    """
    Clean up
    """
    # Remove results of the experiment
    SsResults.remove_all_results()
    NetperfResults.remove_all_results()
    TcResults.remove_all_results()
    PingResults.remove_all_results()

    kill_processes()

# Helper methods
def _find_start_stop_time_for_ss(src_ns, dst_addr, start_t, stop_t, ss_schedules):
    """
    Find the start time and stop time to run ss command in node `src_ns`
    to a `dst_addr`

    Args:
        src_ns: Node
            ss run from `src_ns`
        dst_addr: Address
            Destination address
        start_t: int
            Start time of ss command
        stop_t: int
            Stop time of ss command
        ss_schedules: list
            List with ss command schedules

    Returns:
        List: Updated ss_schedules
    """
    if (src_ns, dst_addr) not in ss_schedules:
        ss_schedules[(src_ns, dst_addr)] = (start_t, stop_t)
    else:
        (min_start, max_stop) = ss_schedules[(src_ns, dst_addr)]
        ss_schedules[(src_ns, dst_addr)] = (min(min_start, start_t), max(max_stop, stop_t))

    return ss_schedules



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""API related to interfaces in topology"""

from .address import Address
from .. import engine
from .id_generator import IdGen
from ..topology_map import TopologyMap
from . import traffic_control
from .. import config

# TODO: Improve this module such that the below pylint disables are no
# longer required

#pylint: disable=too-many-instance-attributes
#pylint: disable=protected-access
#pylint: disable=too-few-public-methods

class Interface:
    """
    Abstraction for an network interface.

    Attributes
    ----------
    name: str
        User given name for the interface
    id: str
        This value is used by `engine` to create emulated interface
        entity
    node: Node
        `Node` which contains this `Interface`
    address: str/Address
        IP address assigned to this interface
    """

    def __init__(self, interface_name):
        """
        name is only used to display it to the user
        id is the actual backend name
        namespace is of namespace class and not a string
        pair is another object of Interface class to which this is connected

        set_structure tells us if ifb, a default qdisc, netem and htb are added
        ifb is an object of interface class which tells the ifb associated with this interface

        qdisc_list, class_list, filter_list gives a list of all those as respective classes

        TODO: Add parameter list
        """

        # TODO: name and address should be the only public members
        self._name = interface_name
        self._id = IdGen.get_id(interface_name)

        self._address = None
        self._node = None
        self._pair = None

        self.set_structure = False
        self.ifb = None

        # TODO: These are not rightly updated
        # set_delay and set_bandwidth invoke the
        # engine function directly
        self.qdisc_list = []
        self.class_list = []
        self.filter_list = []

    @property
    def name(self):
        """Getter for name"""
        return self._name

    @property
    def id(self):
        """Getter for id"""
        return self._id

    @property
    def pair(self):
        """
        Get other pair for this interface (assuming veth)
        """
        return self._pair

    @pair.setter
    def pair(self, interface):
        """
        Setter for the other end of the interface that it is connected to

        Parameters
        ----------
        interface : Interface
            The interface to which this interface is connected to
        """
        self._pair = interface

    @property
    def node(self):
        """
        Getter for the `Node` associated
        with the interface
        """
        return self._node

    @node.setter
    def node(self, node):
        """
        Setter for the `Node` associated
        with the interface

        Parameters
        ----------
        node : Node
            The node where the interface is to be installed
        """
        self._node = node

    @property
    def subnet(self):
        """Getter for the subnet to which the address belongs to"""
        return self.address.get_subnet()

    @property
    def address(self):
        """
        Getter for the address associated
        with the interface
        """
        return self._address

    @address.setter
    def address(self, address):
        """
        Assigns ip adress to an interface

        Parameters
        ----------
        address : Address or str
            ip address to be assigned to the interface
        """
        if isinstance(address, str):
            address = Address(address)

        if self.node is not None:
            engine.assign_ip(self.node.id, self.id, address.get_addr())
            self._address = address
        else:
            # TODO: Create our own error class
            raise NotImplementedError(
                'You should assign the interface to node or router before assigning address to it.')

    def get_address(self):
        """
        Getter for the address associated
        with the interface

        *NOTE*: Maintained since mentioned in NeST paper.
        """
        return self.address

    def set_address(self, address):
        """
        Assigns ip adress to an interface

        Parameters
        ----------
        address : Address or str
            ip address to be assigned to the interface

        *NOTE*: Maintained since mentioned in NeST paper.
        """
        self.address = address

    def set_mode(self, mode):
        """
        Changes the mode of the interface

        Parameters
        ----------
        mode : string
            interface mode to be set
        """
        if mode in ('UP', 'DOWN'):
            if self.node is not None:
                engine.set_interface_mode(
                    self.node.id, self.id, mode.lower())
            else:
                # TODO: Create our own error class
                raise NotImplementedError(
                    'You should assign the interface to node or router before setting it\'s mode')
        else:
            raise ValueError(f'{mode} is not a valid mode (it has to be either "UP" or "DOWN")')

    def get_qdisc(self):
        """
        Note that this is the qdisc set inside
        the IFB.
        """
        if self.ifb is not None:
            for qdisc in self.ifb.qdisc_list:
                if qdisc.parent == '1:1' and qdisc.handle == '11:':
                    return qdisc
        return None

    def add_qdisc(self, qdisc, parent='root', handle='', **kwargs):
        """
        Add a qdisc (Queueing Discipline) to this interface

        Parameters
        ----------
        qdisc : string
            The qdisc which needs to be added to the interface
        dev : Interface class
            The interface to which the qdisc is to be added
        parent : string
            id of the parent class in major:minor form(optional) (Default value = 'root')
        handle : string
            id of the filter (Default value = '')
        """
        self.qdisc_list.append(traffic_control.Qdisc(
            self.node.id, self.id, qdisc, parent, handle, **kwargs))

        # Add qdisc to TopologyMap
        TopologyMap.add_qdisc(self.node.id, self.id, qdisc, handle, parent=parent)

        return self.qdisc_list[-1]

    def delete_qdisc(self, handle):
        """
        Delete qdisc (Queueing Discipline) from this interface

        Parameters
        ----------
        handle : string
            Handle of the qdisc to be deleted
        """
        # TODO: Handle this better by using the destructor in traffic-control
        counter = 0
        for qdisc in self.qdisc_list:
            if qdisc.handle == handle:
                engine.delete_qdisc(qdisc.namespace_id,
                                    qdisc.dev_id, qdisc.parent, qdisc.handle)
                TopologyMap.delete_qdisc(self.node.id, self.id, handle)
                self.qdisc_list.pop(counter)
                break
            counter += 1

    def add_class(self, qdisc, parent='root', classid='', **kwargs):
        """
        Create an object that represents a class

        Parameters
        ----------
        qdisc : string
            The qdisc which needs to be added to the interface
        parent : string
            id of the parent class in major:minor form(optional) (Default value = 'root')
        classid : string
            id of the class (Default value = '')
        """
        self.class_list.append(traffic_control.Class(
            self.node.id, self.id, qdisc, parent, classid, **kwargs))

        return self.class_list[-1]

    #pylint: disable=too-many-arguments
    def add_filter(self, priority, filtertype, flowid, protocol='ip', parent='root',
                   handle='', **kwargs):
        """
        Design a Filter to assign to a Class or Qdisc

        Parameters
        ----------
        protocol : string
            protocol used (Default value = 'ip')
        priority : int
            priority of the filter
        filtertype : string
            one of the available filters
        flowid : Class
            classid of the class where the traffic is enqueued
            if the traffic passes the filter
        parent : string
            id of the parent class in major:minor form(optional) (Default value = 'root')
        handle : string
            id of the filter (Default value = '')
        filter : dictionary
            filter parameters
        """
        # TODO: Verify type of parameters
        # TODO: Reduce arguements to the engine functions by finding parent and handle automatically

        self.filter_list.append(
            traffic_control.Filter(
                self.node.id, self.id, protocol, priority, filtertype,
                flowid, parent, handle, **kwargs))

        return self.filter_list[-1]

    def _create_and_mirred_to_ifb(self, dev_name):
        """
        Creates a IFB for the interface so that a Qdisc can be
        installed on it
        Mirrors packets to be sent out of the interface first to
        itself (IFB)
        Assumes the interface has already invoked _set_structure()

        Parameters
        ----------
        dev_name : string
            The interface to which the ifb was added
        """
        self.ifb = Interface('ifb-' + dev_name)
        engine.create_ifb(self.ifb.id)

        # Add ifb to namespace
        self.node._add_interface(self.ifb)

        # Set interface up
        self.ifb.set_mode('UP')

        default_route = {
            'default': '1'
        }

        # TODO: find how to set a good bandwitdh
        default_bandwidth = {
            'rate': config.get_value('default_bandwidth')
        }

        # TODO: Standardize this; seems like arbitrary handle values
        # were chosen.
        self.ifb.add_qdisc('htb', 'root', '1:', **default_route)
        self.ifb.add_class('htb', '1:', '1:1', **default_bandwidth)
        self.ifb.add_qdisc('pfifo', '1:1', '11:')

        action_redirect = {
            'match': 'u32 0 0',  # from man page examples
            'action': 'mirred',
            'egress': 'redirect',
            'dev': self.ifb.id
        }

        # NOTE: Use Filter API
        engine.add_filter(self.node.id, self.id, 'ip', '1', 'u32', parent='1:', **action_redirect)

    def _set_structure(self):
        """
        Sets a proper sturcture to the interface by creating htb class
        with default bandwidth and a netem qdisc as a child.
        (default bandwidth = 1024mbit)
        """
        self.set_structure = True

        default_route = {
            'default': '1'
        }

        self.add_qdisc('htb', 'root', '1:', **default_route)

        # TODO: find how to set a good bandwitdh
        default_bandwidth = {
            'rate': config.get_value('default_bandwidth')
        }

        self.add_class('htb', '1:', '1:1', **default_bandwidth)

        self.add_qdisc('netem', '1:1', '11:')

    def set_bandwidth(self, min_rate):
        """
        Sets a minimum bandwidth for the interface
        It is done by adding a htb qdisc and a rate parameter to the class

        Parameters
        ----------
        min_rate : string
            The minimum rate that has to be set in kbit
        """
        # TODO: Check if there exists a delay and if exists, make sure it is handled in the right
        # way
        # TODO: Check if this is a redundant condition
        # TODO: Let user set the unit

        if self.set_structure is False:
            self._set_structure()

        min_bandwidth_parameter = {
            'rate': min_rate
        }

        # TODO: Check the created API
        # TODO: This should be handled by self.change_class
        engine.change_class(self.node.id, self.id, '1:', 'htb', '1:1', **min_bandwidth_parameter)

    def set_delay(self, delay):
        """
        Adds a delay to the link between two namespaces
        It is done by adding a delay in the interface

        Parameters
        ----------
        delay : str
            The delay to be added
        """
        # TODO: It is not intuitive to add delay to an interface
        # TODO: Make adding delay possible without bandwidth being set
        # TODO: Check if this is a redundant condition
        # TODO: Let user set the unit

        if self.set_structure is False:
            self._set_structure()

        delay_parameter = {
            'delay': delay
        }

        # TODO: This should be handled by self.change_qdisc
        # It could lead to a potential bug!
        engine.change_qdisc(self.node.id, self.id, 'netem', '1:1', '11:', **delay_parameter)

    def set_qdisc(self, qdisc, bandwidth, **kwargs):
        """
        Adds the Queueing algorithm to the interface

        Parameters
        ----------
        qdisc : string
            The Queueing algorithm to be added
        bandwidth :
            Link bandwidth
        """
        # TODO: Don't use this API directly. If it used,
        # then the bandwidth should be same as link bandwidth
        # (A temporary bug fix is causing this issue. Look for
        # a permanent solution)
        # TODO: Check if this is a redundant condition

        if self.set_structure is False:
            self._set_structure()

        if self.ifb is None:
            self._create_and_mirred_to_ifb(self.name)

        min_bandwidth_parameter = {
            'rate': bandwidth
        }

        engine.change_class(self.node.id, self.ifb.id,
                            '1:', 'htb', '1:1', **min_bandwidth_parameter)

        self.ifb.delete_qdisc('11:')
        self.ifb.add_qdisc(qdisc, '1:1', '11:', **kwargs)

    def set_attributes(self, bandwidth, delay, qdisc=None, **kwargs):
        """
        Add attributes bandwidth, delay and qdisc to interface

        Parameters
        ----------
        bandwidth :
            Packet outgoing rate
        delay : string
            Delay before packet is sent out
        qdisc : string
            The Queueing algorithm to be added to interface
            (Default value = None)
        """

        self.set_bandwidth(bandwidth)
        self.set_delay(delay)

        if qdisc is not None:
            self.set_qdisc(qdisc, bandwidth, **kwargs)

    def __repr__(self):
        classname = self.__class__.__name__
        return f'{classname}({self.name!r})'

class Veth:
    """Handle creation of Veth pairs"""

    def __init__(self, interface1_name, interface2_name):
        """
        Constructor to create a veth pair between
        `interface1_name` and `interface2_name`
        """

        self.interface1 = Interface(interface1_name)
        self.interface2 = Interface(interface2_name)

        self.interface1.pair = self.interface2
        self.interface2.pair = self.interface1

        # Create the veth
        engine.create_veth(self.interface1.id, self.interface2.id)

    def _get_interfaces(self):
        """Get tuple of endpoint interfaces"""

        return (self.interface1, self.interface2)


def connect(ns1, ns2, interface1_name='', interface2_name=''):
    """
    Connects two nodes `ns1` and `ns2`
    """
    # Create 2 interfaces

    if interface1_name == '' and interface2_name == '':
        connections = _number_of_connections(ns1, ns2)
        interface1_name = ns1.name + '-' + ns2.name + '-' + str(connections)
        interface2_name = ns2.name + '-' + ns1.name + '-' + str(connections)

    veth = Veth(interface1_name, interface2_name)
    (int1, int2) = veth._get_interfaces()

    ns1._add_interface(int1)
    ns2._add_interface(int2)

    # Set the proper structure for the veth
    int1._set_structure()
    int2._set_structure()

    int1.set_mode('UP')
    int2.set_mode('UP')

    return (int1, int2)


def _number_of_connections(ns1, ns2):
    """
    Return number of connections between the two namespaces
    """

    connections = 0

    if len(ns1.interfaces) > len(ns2.interfaces):
        ns1, ns2 = ns2, ns1

    for interface in ns1.interfaces:
        if interface.pair.node == ns2:
            connections = connections + 1

    return connections
